import io
from typing import IO, TYPE_CHECKING, Any, Dict, List, Optional

import numpy as np

from .fiber import Fiber

if TYPE_CHECKING:
    from .nodeset import PointNodeset, LineNodeset, SurfaceNodeset, VolumeNodeset


class Node:
    """
    Class that holds all information of nodes like coords, fibers, nodesets (and additional data)
    """

    def __init__(self, coords: np.ndarray = np.zeros((3))):
        """
        Initialize node at the coordinades coords

        Args:
            coords: np.array((3)) Coordinates of the node
        """
        self.id: Optional[int] = None
        self.coords: np.ndarray = coords
        self.fibers: Dict[str, Fiber] = {}

        self.pointnodesets: List[PointNodeset] = []
        self.linenodesets: List[LineNodeset] = []
        self.surfacenodesets: List[SurfaceNodeset] = []
        self.volumenodesets: List[VolumeNodeset] = []
        self.data: Dict[str, Any] = {}

    def reset(self) -> None:
        """
        Sets the id to None
        """
        self.id = None

    def get_line(self) -> str:
        """
        Returns the line definition in the dat file
        """
        dest = io.StringIO()
        if len(self.fibers) > 0:
            dest.write("FNODE")
        else:
            dest.write("NODE")

        if self.id is None:
            raise RuntimeError("You have to compute ids before writing")

        dest.write(
            " {0} COORD {1}".format(self.id, " ".join([str(i) for i in self.coords]))
        )

        for k, f in self.fibers.items():
            dest.write(" ")
            f.write(dest, k)

        return dest.getvalue()

    def write(self, dest: IO) -> None:
        """
        Writes the corresponding line in to the stream variable

        Args:
            dest: stream variable where to write the line
        """
        if len(self.fibers) > 0:
            dest.write("FNODE")
        else:
            dest.write("NODE")

        if self.id is None:
            raise RuntimeError("You have to compute ids before writing")

        dest.write(
            " {0} COORD {1}".format(self.id, " ".join([str(i) for i in self.coords]))
        )

        for k, f in self.fibers.items():
            dest.write(" ")
            f.write(dest, k)

        dest.write("\n")



import os
import unittest
from typing import List

import lnmmeshio
import numpy as np

script_dir = os.path.dirname(os.path.realpath(__file__))


class TestElementFactory(unittest.TestCase):
    def setUp(self):
        pass

    @staticmethod
    def get_nodes(num: int) -> List[lnmmeshio.Node]:
        return [lnmmeshio.Node(np.array([0, 0, 0])) for i in range(num)]

    def test_element_factory(self):
        self.assert_element(lnmmeshio.Tet10, 10)
        self.assert_element(lnmmeshio.Tet4, 4)
        self.assert_element(lnmmeshio.Hex8, 8)
        self.assert_element(lnmmeshio.Hex20, 20)
        self.assert_element(lnmmeshio.Hex27, 27)

    def assert_element(self, cls, numnodes: int):
        ele = lnmmeshio.element.factory(
            cls.ShapeName, TestElementFactory.get_nodes(numnodes)
        )
        self.assertIsInstance(ele, cls)



#
# This is the interface between the meshio package and our Discretization description
# May not work perfectly
#
from typing import Any, Dict, List, Optional, Tuple

import meshio
import numpy as np
from tqdm import tqdm

from .discretization import Discretization, Node
from .element.line2 import Line2
from .element.line3 import Line3
from .element.parse_element import create_element
from .element.quad4 import Quad4
from .element.tri3 import Tri3
from .element.tri6 import Tri6
from .nodeset import (
    LineNodeset,
    LineNodesetBuilder,
    PointNodeset,
    PointNodesetBuilder,
    SurfaceNodeset,
    SurfaceNodesetBuilder,
    VolumeNodeset,
    VolumeNodesetBuilder,
)

cell_nodes = {
    "vertex": 1,
    "line": 2,
    "line3": 3,
    "triangle": 3,
    "triangle6": 6,
    "tetra": 4,
    "tetra10": 10,
    "pyramid": 5,
    "hexahedron": 8,
    "hexahedron20": 20,
    "hexahedron27": 27,
    "wedge": 6,
    "quad": 4,
    "quad9": 9,
    "vertex": 1,
}

cell_disc_eles = {
    "vertex": "UNKNOWN",
    "line": "UNKNOWN",
    "line3": "UNKNOWN",
    "triangle": "UNKNOWN",
    "triangle6": "UNKNOWN",
    "tetra": "SOLID",
    "tetra10": "SOLID",
    "pyramid": "SOLID",
    "hexahedron": "SOLID",
    "hexahedron20": "SOLID",
    "hexahedron27": "SOLID",
    "wedge": "UNKNOWN",
    "quad": "UNKNOWN",
    "quad9": "UNKNOWN",
}

cell_disc_shape = {
    "vertex": "VERTEX1",
    "line": "LINE2",
    "line3": "LINE3",
    "triangle": "TRI3",
    "triangle6": "TRI6",
    "tetra": "TET4",
    "tetra10": "TET10",
    "pyramid": "PYRAMID5",
    "hexahedron": "HEX8",
    "hexahedron20": "HEX20",
    "hexahedron27": "HEX27",
    "wedge": "WEDGE6",
    "quad": "QUAD4",
    "quad9": "QUAD9",
}

_cell_data_id_names = ["medit:ref", "gmsh:geometrical"]

# invert cell_disc_shape
disc_shape_cell = {v: k for k, v in cell_disc_shape.items()}

cell_to_dim = {
    "vertex": 0,
    "line": 1,
    "line3": 1,
    "triangle": 2,
    "triangle6": 2,
    "tetra": 3,
    "tetra10": 3,
    "pyramid": 3,
    "hexahedron": 3,
    "hexahedron20": 3,
    "hexahedron27": 3,
    "wedge": 3,
    "quad": 2,
    "quad9": 2,
}

# meshio used vtk node ordering
# with this array, we reorder the nodes
ele_node_order_vtk2baci = {
    "VERTEX1": list(range(0, 1)),
    "LINE2": list(range(0, 2)),
    "LINE3": list(range(0, 2)),
    "TRI3": list(range(0, 3)),
    "TRI6": list(range(0, 6)),
    "TET4": list(range(0, 4)),
    "TET10": list(range(0, 10)),
    "PYRAMID5": list(range(0, 5)),
    "HEX8": list(range(0, 8)),
    "HEX20": [
        0,
        1,
        2,
        3,
        4,
        5,
        6,
        7,
        8,
        9,
        10,
        11,
        16,
        17,
        18,
        19,
        12,
        13,
        14,
        15,
    ],  # verify ?
    "HEX27": [
        0,
        1,
        2,
        3,
        4,
        5,
        6,
        7,
        8,
        9,
        10,
        11,
        12,
        13,
        14,
        15,
        16,
        17,
        18,
        19,
        21,
        25,
        24,
        26,
        23,
        22,
        20,
    ],
    "WEDGE6": list(range(0, 6)),
    "QUAD4": list(range(0, 4)),
    "QUAD9": list(range(0, 9)),
}


def mesh2Discretization(mesh: meshio.Mesh) -> Discretization:
    # create empty discretization
    disc = Discretization()

    # mesh.points -> disc.nodes
    # mesh.cells -> disc.elements[Element.FieldTypeStructure]
    # node_sets -> are stored in disc.nodes

    # create Nodes from points
    for i, coord in tqdm(enumerate(mesh.points), desc="Create nodes"):
        n = Node(coord)

        for key, v in mesh.point_data.items():
            n.data[key] = v[i]

        disc.nodes.append(n)

    # get the maximum element dimension, which is the dimension of the mesh
    maxdim = max([cell_to_dim[cell_block.type] for cell_block in mesh.cells])

    # create Elements from cells
    disc.elements.structure = []
    pointnsbuilder = PointNodesetBuilder()
    linensbuilder = LineNodesetBuilder()
    surfnsbuilder = SurfaceNodesetBuilder()
    volumensbuilder = VolumeNodesetBuilder()

    for cellgroupid, cellblock in enumerate(mesh.cells):
        if cellblock.type not in cell_nodes:
            raise Exception(
                "The cell type {0} is currently not implemented".format(cellblock.type)
            )

        for cellid, cell in enumerate(cellblock.data):
            # read nodes that belong to the cells
            nodes = []

            # in case of a vertex, cell is just an integer
            cell = np.array(cell).reshape((-1))

            for nodeids in cell:
                nodes.append(disc.nodes[nodeids])

            # cells with the same dimension as the mesh dimensions are normal cells
            # cells with a lower dimension are treated as surface, line or node
            # definitions
            element_dim = cell_to_dim[cellblock.type]
            if maxdim == element_dim:
                # this is a normal element
                ele = create_element(
                    cell_disc_eles[cellblock.type],
                    cell_disc_shape[cellblock.type],
                    # also need to reorder nodes in a few element types
                    [
                        nodes[i]
                        for i in ele_node_order_vtk2baci[
                            cell_disc_shape[cellblock.type]
                        ]
                    ],
                )
                disc.elements.structure.append(ele)

                # extract material info from cell data
                matid: Optional[int] = _get_material_from_cell_data(
                    mesh.cell_data, cellgroupid, cellid
                )
                if matid == None:
                    matid = 1
                ele.options["MAT"] = matid

                # extract cell data
                for key, data in mesh.cell_data.items():
                    ele.data[key] = data[cellgroupid][cellid]

                # explicitly append the cell group id (block id) to the cell data
                ele.data["GROUP_ID"] = cellgroupid

                # create surface-nodesets
                if maxdim == 2 and len(mesh.cell_data) > 0:
                    nsid: Optional[int] = _get_nodesetid_from_cell_data(
                        mesh.cell_data, cellgroupid, cellid
                    )

                    if nsid is not None:
                        for node in nodes:
                            surfnsbuilder.add(node, nsid)

            else:
                # this is a lower-dimensional element
                # treat as surface-, line- or point-nodeset definition

                # extract nodeset info from cell data
                nsid: Optional[int] = _get_nodesetid_from_cell_data(
                    mesh.cell_data, cellgroupid, cellid
                )

                if nsid is not None:
                    for node in nodes:
                        if element_dim == 0:
                            pointnsbuilder.add(node, nsid)
                        elif element_dim == 1:
                            linensbuilder.add(node, nsid)
                        elif element_dim == 2:
                            surfnsbuilder.add(node, nsid)

    # go through nodesets
    for name, nodes in mesh.point_sets.items():
        if "volume" in name:
            nsid = volumensbuilder.get_unused_id()
            for n in nodes:
                volumensbuilder.add(disc.nodes[n], nsid)
        if "surface" in name:
            nsid = surfnsbuilder.get_unused_id()
            for n in nodes:
                surfnsbuilder.add(disc.nodes[n], nsid)
        elif "line" in name:
            nsid = linensbuilder.get_unused_id()
            for n in nodes:
                linensbuilder.add(disc.nodes[n], nsid)
        elif "point" in name:
            nsid = pointnsbuilder.get_unused_id()
            for n in nodes:
                pointnsbuilder.add(disc.nodes[n], nsid)

    disc.volumenodesets = volumensbuilder.finalize()
    disc.pointnodesets = pointnsbuilder.finalize()
    disc.linenodesets = linensbuilder.finalize()
    disc.surfacenodesets = surfnsbuilder.finalize()

    disc.finalize()
    return disc


def discretization2mesh(dis: Discretization) -> meshio.Mesh:
    dis.compute_ids(zero_based=True)

    points = dis.get_node_coords()

    point_data = {}

    for i, n in enumerate(dis.nodes):
        for k, v in n.data.items():
            if k not in point_data:
                point_data[k] = np.zeros(
                    tuple([len(dis.nodes)] + list(np.array(v).shape))
                )

            point_data[k][i] = v

    cells = []
    cell_data = {}

    id_to_index = {}

    # store cells
    for eletype in dis.elements.values():
        for ele in eletype:
            celltype = disc_shape_cell[ele.shape]

            newgroup = False

            if len(cells) == 0 or cells[-1][0] != celltype:
                cell_nodes = []
                cells.append((celltype, cell_nodes))
                newgroup = True
            else:
                cell_nodes = cells[-1][1]

            cell_nodes.append([n.id for n in ele.nodes])
            id_to_index[ele.id] = (len(cells) - 1, len(cell_nodes) - 1)

    # go again through the elements to store element data
    for eletype in dis.elements.values():
        for ele in eletype:
            block_id, cell_id = id_to_index[ele.id]

            if "MAT" in ele.options:
                if "material" not in cell_data:
                    cell_data["material"] = _create_cell_variable(
                        cells, tuple([1]), int
                    )

                cell_data["material"][block_id][cell_id] = int(
                    ele.options["MAT"][0]
                    if _isiter(ele.options["MAT"])
                    else ele.options["MAT"]
                )

            for variable_name, value in ele.data.items():
                value_reshaped = np.array(value).reshape((-1))

                if variable_name not in cell_data:
                    cell_data[variable_name] = _create_cell_variable(
                        cells, value_reshaped.shape, value_reshaped.dtype
                    )
                cell_data[variable_name][block_id][cell_id] = value_reshaped

    mesh: meshio.Mesh = meshio.Mesh(
        points, cells, cell_data=cell_data, point_data=point_data
    )

    dis.reset()

    return mesh


def _create_cell_variable(
    cells: List[Tuple[str, List[int]]], shape: Tuple[int], dtype
) -> List[np.ndarray]:
    data = []

    for block in cells:
        new_shape = tuple([len(block[1])] + list(shape))
        if len(shape) == 1 and shape[0] == 1:
            new_shape = tuple([len(block[1])])
        data.append(np.zeros(new_shape, dtype=dtype))

    return data


def _get_material_from_cell_data(
    celldata: Dict[str, List[List[Any]]], cellgroupid: int, cellid: int
) -> Optional[int]:
    for name in _cell_data_id_names:
        if name in celldata:
            return int(celldata[name][cellgroupid][cellid])

    return None


def _get_nodesetid_from_cell_data(
    celldata: Dict[str, List[List[Any]]], cellgroupid: int, cellid: int
) -> Optional[int]:
    for name in _cell_data_id_names:
        if name in celldata:
            return int(celldata[name][cellgroupid][cellid])

    return None


def _isiter(list: Any) -> bool:
    try:
        iter(list)
    except TypeError:
        return False

    return True



import math
from typing import List

import numpy as np

from ..node import Node
from .element import ElementTri
from .line3 import Line3


class Tri6(ElementTri):
    """
    Implementation of a tri6 element
    """

    ShapeName: str = "TRI6"

    def __init__(self, el_type: str, nodes: List[Node]):
        """
        Base constructor of a tri6 element
        """
        super(Tri6, self).__init__(el_type, Tri6.ShapeName, nodes)

        if len(nodes) != self.get_num_nodes():
            raise RuntimeError(
                "You tried to created a TRI6 element with {0} nodes".format(len(nodes))
            )

    @classmethod
    def get_num_nodes(cls) -> int:
        """
        Get number of nodes of a TRI6 element

        Returns:
            Number of nodes of a TRI6 element = 6
        """
        return 6

    def get_faces(self) -> List["Tri6"]:
        """
        A TRI6 element does not contain any faces

        Returns:
            List of faces
        """
        return [self]

    def get_edges(self) -> List[Line3]:
        """
        Returns the list of all edges

        Returns:
            List of edges
        """
        return [
            Line3(None, [self.nodes[0], self.nodes[1], self.nodes[3]]),
            Line3(None, [self.nodes[1], self.nodes[2], self.nodes[4]]),
            Line3(None, [self.nodes[2], self.nodes[0], self.nodes[5]]),
        ]

    def integrate_xi(self, integrand, numgp) -> np.ndarray:
        """
        Integrates the integrand over the element

        Args:
            integrand: integrand as a function of xi
            numgp: Number of integration points to be used
        """
        result = np.array(0.0)

        intpoints: np.ndarray = ElementTri.int_points(numgp)
        intweights: np.ndarray = ElementTri.int_weights(numgp)

        coords = np.array([n.coords for n in self.nodes])

        for xi, w in zip(intpoints, intweights):
            J = np.matmul(Tri6.shape_fcns_derivs(xi), coords)
            G = np.matmul(J, J.T)
            detA = math.sqrt(np.linalg.det(G))
            result += w * detA * integrand(xi)

        return result

    def integrate(self, integrand, numgp) -> np.ndarray:
        """
        Integrates the integrand over the element

        Args:
            integrand: integrand as a function of x
            numgp: Number of integration points to be used
        """
        result = np.array(0.0)

        intpoints: np.ndarray = ElementTri.int_points(numgp)
        intweights: np.ndarray = ElementTri.int_weights(numgp)

        coords = np.array([n.coords for n in self.nodes])

        for xi, w in zip(intpoints, intweights):
            x = np.matmul(Tri6.shape_fcns(xi), coords)
            J = np.matmul(Tri6.shape_fcns_derivs(xi), coords)
            G = np.matmul(J, J.T)
            detA = math.sqrt(np.linalg.det(G))
            result += w * detA * integrand(x)

        return result

    def get_normal(self, xi):
        coords = np.array([n.coords for n in self.nodes])
        J = np.matmul(Tri6.shape_fcns_derivs(xi), coords)

        n = np.cross(J[0, :], J[1, :])
        return n / np.linalg.norm(n)

    @staticmethod
    def shape_fcns(xi):
        """
        Returns the value of the shape functions at the local coordinate xi
        """
        r = xi[0]
        s = xi[1]

        t1 = 1.0 - r - s
        t2 = r
        t3 = s

        return np.array(
            [
                t1 * (2.0 * t1 - 1.0),
                t2 * (2.0 * t2 - 1.0),
                t3 * (2.0 * t3 - 1.0),
                4.0 * t2 * t1,
                4.0 * t2 * t3,
                4.0 * t3 * t1,
            ]
        )

    @staticmethod
    def shape_fcns_derivs(xi):
        """
        Returns the value of the derivatives of the shape functions with respect to the local coordinates at the local coordinate xi

        +-                                  -+
        |  dN_1 / dxi_1   dN_2 / dxi_1, ...  |
        |  dN_1 / dxi_2   dN_2 / dxi_2, ...  |
        +-                                  -+
        """
        t1 = 1.0 - xi[0] - xi[1]
        t2 = xi[0]
        t3 = xi[1]

        return np.array(
            [
                [
                    -(4.0 * t1 - 1),
                    4.0 * t2 - 1,
                    0.0,
                    4.0 * t1 - 4.0 * t2,
                    4.0 * t3,
                    -4.0 * t3,
                ],
                [
                    -(4.0 * t1 - 1),
                    0.0,
                    4.0 * t3 - 1,
                    -4.0 * t2,
                    4 * t2,
                    4.0 * t1 - 4.0 * t3,
                ],
            ]
        )



import re
from typing import List, Optional

from ..fiber import Fiber
from ..ioutils import (
    read_key_values,
    read_option_items,
    write_option,
    write_option_list,
    write_title,
)
from ..node import Node
from .element import Element
from .hex8 import Hex8
from .hex20 import Hex20
from .hex27 import Hex27
from .line2 import Line2
from .line3 import Line3
from .quad4 import Quad4
from .quad8 import Quad8
from .quad9 import Quad9
from .tet4 import Tet4
from .tet10 import Tet10
from .tri3 import Tri3
from .tri6 import Tri6
from .vertex import Vertex

RegExEle = re.compile(r"^[ ]*([0-9]+)[ ]+(\S+)[ ]+(\S+)[ ]+")


def create_element(
    ele_type: str, ele_shape: str, ele_nodes: List[Node], throw_if_unknown=False
) -> Element:
    """
    Creates an element from a given element shape

    Args:
        elename: name of the element (e.g. SOLID)
        shape: str shape of the element (e.g. HEX8)
        nodes: nodes of the element
        throw_if_unknown: bool Throw an exception, if element type is not known

    Return:
        Element of the specific type
    """

    ele: Optional[Element] = None
    if ele_shape == Vertex.ShapeName:
        ele = Vertex(ele_type, ele_nodes)
    elif ele_shape == Line2.ShapeName:
        ele = Line2(ele_type, ele_nodes)
    elif ele_shape == Line3.ShapeName:
        ele = Line3(ele_type, ele_nodes)
    elif ele_shape == Tri3.ShapeName:
        ele = Tri3(ele_type, ele_nodes)
    elif ele_shape == Tri6.ShapeName:
        ele = Tri6(ele_type, ele_nodes)
    elif ele_shape == Quad4.ShapeName:
        ele = Quad4(ele_type, ele_nodes)
    elif ele_shape == Quad8.ShapeName:
        ele = Quad8(ele_type, ele_nodes)
    elif ele_shape == Quad9.ShapeName:
        ele = Quad9(ele_type, ele_nodes)
    elif ele_shape == Tet4.ShapeName:
        ele = Tet4(ele_type, ele_nodes)
    elif ele_shape == Tet10.ShapeName:
        ele = Tet10(ele_type, ele_nodes)
    elif ele_shape == Hex8.ShapeName:
        ele = Hex8(ele_type, ele_nodes)
    elif ele_shape == Hex20.ShapeName:
        ele = Hex20(ele_type, ele_nodes)
    elif ele_shape == Hex27.ShapeName:
        ele = Hex27(ele_type, ele_nodes)
    else:
        if throw_if_unknown:
            raise RuntimeError("The element type {0} is unknown".format(ele_shape))
        ele = Element(ele_type, ele_shape, ele_nodes)

    return ele


def parse(line: str, nodes: List[Node], throw_if_unknown=False):
    """
    Parses the element and returns an instance of an appropriate element type

    Args:
        line: linedefinition of the element
        nodes: List of nodes

    Returns:
        An instance of the properly instantianted element
    """
    line = line.split("//", 1)[0]
    # parse ele id, type and shape
    ele_match = RegExEle.search(line)
    if not ele_match:
        return None

    ele_id = int(ele_match.group(1))
    ele_type = ele_match.group(2)
    ele_shape = ele_match.group(3)

    node_ids_str, span = read_option_items(
        line, ele_shape, Element.num_nodes_by_shape(ele_shape)
    )
    ele_nodes = [nodes[int(i) - 1] for i in node_ids_str]

    ele = create_element(
        ele_type, ele_shape, ele_nodes, throw_if_unknown=throw_if_unknown
    )

    ele.id = ele_id

    # read fibers
    ele.fibers = Fiber.parse_fibers(line)

    # read remaining options
    # assume only one value per option, which must not be the case in general
    line = line[span[1] :]

    for key, values in read_key_values(
        line, lambda key: 3 if Fiber.is_fiber_type(key) else 1
    ):
        if len(values) == 1:
            ele.options[key] = values[0]
        else:
            ele.options[key] = values

    return ele



from typing import IO, TYPE_CHECKING, Dict, List, Set

import numpy as np
import yaml
from tqdm import tqdm

from .element.element import Element1D, Element2D, Element3D
from .element.element_container import ElementContainer
from .fiber import Fiber
from .ioutils import line_option, read_option_item, read_option_items, write_title
from .node import Node
from .nodeset import LineNodeset, PointNodeset, SurfaceNodeset, VolumeNodeset


class Discretization:
    """
    This class holds the discretization, consisting out of nodes and elements. The nodes and
    elements itself hold their data (coords, element type, ...)
    """

    def __init__(self):
        """
        Initialize Discretization class with empty nodes and zero elements
        """
        self.nodes: List[Node] = []
        self.elements: ElementContainer = ElementContainer()

        # initialize nodesets
        self.pointnodesets: List[PointNodeset] = []
        self.linenodesets: List[LineNodeset] = []
        self.surfacenodesets: List[SurfaceNodeset] = []
        self.volumenodesets: List[VolumeNodeset] = []

    def compute_ids(self, zero_based: bool) -> None:
        """
        Computes the ids of the elements and nodes.

        Args:
            zero_based: If true, the first node id is 0, otherwise 1
        """

        id: int = 0 if zero_based else 1
        for node in self.nodes:
            node.id = id
            id += 1

        id: int = 0 if zero_based else 1
        for ele_i in self.elements.values():
            for ele in ele_i:
                ele.id = id
                id += 1

        id: int = 0 if zero_based else 1
        for ns in self.pointnodesets:
            ns.id = id
            id += 1

        id: int = 0 if zero_based else 1
        for ns in self.linenodesets:
            ns.id = id
            id += 1

        id: int = 0 if zero_based else 1
        for ns in self.surfacenodesets:
            ns.id = id
            id += 1

        id: int = 0 if zero_based else 1
        for ns in self.volumenodesets:
            ns.id = id
            id += 1

    def reset(self) -> None:
        """
        Resets the computed ids
        """
        for node in self.nodes:
            node.reset()

        for ele_i in self.elements.values():
            for ele in ele_i:
                ele.reset()

        for ns in self.pointnodesets:
            ns.reset()
        for ns in self.linenodesets:
            ns.reset()
        for ns in self.surfacenodesets:
            ns.reset()
        for ns in self.volumenodesets:
            ns.reset()

    def get_node_coords(self) -> np.ndarray:
        """
        Returns an np.array((num_node, 3)) with the coordinates of each node
        """
        arr: np.ndarray = np.zeros((len(self.nodes), 3))

        for i, node in enumerate(self.nodes):
            arr[i, :] = node.coords
            i += 1

        return arr

    def get_dline_elements(self, id: int) -> List[Element1D]:
        """
        Returns a list of line elements that belong to a dline
        """
        edge_elements: List[Element1D] = []
        added_edges: Set[str] = set()

        self.compute_ids(True)
        nodeset_ids: Set[int] = set([n.id for n in self.linenodesets[id]])

        for ele in self.elements.structure:
            for edge in ele.get_edges():
                node_ids = [n.id for n in edge.nodes]
                if all([node_id in nodeset_ids for node_id in node_ids]):
                    edge_id = "/".join(sorted([str(nid) for nid in node_ids]))
                    if edge_id not in added_edges:
                        edge_elements.append(edge)
                        added_edges.add(edge_id)

        return edge_elements

    def get_dsurf_elements(self, id: int) -> List[Element2D]:
        """
        Returns a list of surface elements that belong to a dsurf
        """
        face_elements: List[Element2D] = []
        added_faces: Set[str] = set()

        self.compute_ids(True)
        nodeset_ids: Set[int] = set([n.id for n in self.surfacenodesets[id]])
        for element_container in self.elements.values():
            for ele in element_container:
                for face in ele.get_faces():
                    node_ids = [n.id for n in face.nodes]
                    if all([node_id in nodeset_ids for node_id in node_ids]):
                        face_id = "/".join(sorted([str(nid) for nid in node_ids]))
                        if face_id not in added_faces:
                            face_elements.append(face)
                            added_faces.add(face_id)

        return face_elements

    def get_dvol_elements(self, id: int) -> List[Element3D]:
        """
        Returns a list of volume elements that belong to a dvol
        """
        vol_elements: List[Element3D] = []
        added_vols: Set[str] = set()

        self.compute_ids(True)
        nodeset_ids: Set[int] = set([n.id for n in self.volumenodesets[id]])

        for element_container in self.elements.values():
            for ele in element_container:
                if not isinstance(ele, Element3D):
                    continue

                node_ids = [n.id for n in ele.nodes]
                if all([node_id in nodeset_ids for node_id in node_ids]):
                    vol_id = "/".join(sorted([str(nid) for nid in node_ids]))
                    if vol_id not in added_vols:
                        vol_elements.append(ele)
                        added_vols.add(vol_id)
        return vol_elements

    def get_sections(self, out=True) -> Dict[str, List[str]]:
        self.compute_ids(zero_based=False)

        sections = {}
        # write topology
        for ns in [
            self.pointnodesets,
            self.linenodesets,
            self.surfacenodesets,
            self.volumenodesets,
        ]:
            for nsi in tqdm(
                ns, disable=not out, desc="Write Nodeset {0}".format(type(ns))
            ):
                section_name = nsi.get_section()
                if section_name not in sections:
                    sections[section_name] = []

                sections[section_name].extend(nsi.get_lines())

        # write nodes
        nodes = []
        for node in tqdm(self.nodes, disable=not out, desc="Write Nodes"):
            nodes.append(node.get_line())
        sections["NODE COORDS"] = nodes

        # write elements
        sections.update(self.elements.get_sections(out=out))

        return sections

    def write_legacy_dat(self, dest: IO, out: bool = True) -> None:
        """
        Writes the discretization related sections into the stream variable dest in legacy dat format

        Args:
            dest: stream variable (could for example be: with open('file.dat', 'w') as dest: ...)
        """
        sections = self.get_sections(out=out)

        for key, lines in tqdm(
            sections.items(), disable=not out, desc="Write sections"
        ):
            write_title(dest, key)
            for l in lines:
                dest.write("{0}\n".format(l))

    def write_yaml(self, dest: IO, out: bool = True) -> None:
        """
        Writes the discretization related sections into the stream variable dest in 4C yaml format

        Args:
            dest: stream variable (could for example be: with open('file.4C.yaml', 'w') as dest: ...)
        """
        sections = self.get_sections(out=out)
        yaml.dump(sections, dest)

    def finalize(self) -> None:
        """
        Finalizes the discretization by creating internal references
        """

        # remove old nodesets
        for n in self.nodes:
            n.pointnodesets.clear()
            n.linenodesets.clear()
            n.surfacenodesets.clear()
            n.volumenodesets.clear()

        # add point nodesets
        for ns in self.pointnodesets:
            for n in ns:
                n.pointnodesets.append(ns)

        # add line nodesets
        for ns in self.linenodesets:
            for n in ns:
                n.linenodesets.append(ns)

        # add surface nodesets
        for ns in self.surfacenodesets:
            for n in ns:
                n.surfacenodesets.append(ns)

        # add volume nodesets
        for ns in self.volumenodesets:
            for n in ns:
                n.volumenodesets.append(ns)

    @staticmethod
    def read(sections: Dict[str, List[str]], out: bool = False) -> "Discretization":
        """
        Static method that creates the discretizations file from the input lines of a .dat file

        Args:
            sections: Dictionary with header titles as keys and list of lines as value

        Retuns:
            Discretization object
        """
        disc = Discretization()

        # read nodes
        for line in tqdm(sections["NODE COORDS"], disable=not out, desc="Nodes"):
            if "FNODE" in line:
                # this is a fiber node
                nodeid, _ = read_option_item(line, "FNODE")
            else:
                nodeid, _ = read_option_item(line, "NODE")

            if nodeid is None or nodeid == "":
                # this is not a node, probably a comment
                continue

            coords_str, _ = read_option_items(line, "COORD", num=3)

            coords = np.array([float(i) for i in coords_str])

            node = Node(coords=coords)
            disc.nodes.append(node)

            # safety check for integrity of the dat file
            if int(nodeid) != len(disc.nodes):
                raise RuntimeError(
                    "Node ids in dat file have a gap at {0} != {1}!".format(
                        nodeid, len(disc.nodes)
                    )
                )

            # read fibers
            node.fibers = Fiber.parse_fibers(line)

        # read DPOINT topology
        if "DNODE-NODE TOPOLOGY" in sections:
            disc.pointnodesets = PointNodeset.read(
                sections["DNODE-NODE TOPOLOGY"], disc.nodes
            )

        # read DLINE topology
        if "DLINE-NODE TOPOLOGY" in sections:
            disc.linenodesets = LineNodeset.read(
                sections["DLINE-NODE TOPOLOGY"], disc.nodes
            )

        # read DSURF topology
        if "DSURF-NODE TOPOLOGY" in sections:
            disc.surfacenodesets = SurfaceNodeset.read(
                sections["DSURF-NODE TOPOLOGY"], disc.nodes
            )

        # read DVOL topology
        if "DVOL-NODE TOPOLOGY" in sections:
            disc.volumenodesets = VolumeNodeset.read(
                sections["DVOL-NODE TOPOLOGY"], disc.nodes
            )

        # read elements
        disc.elements = ElementContainer.read_element_sections(
            sections, disc.nodes, out=out
        )

        # finalize discretization -> Creates internal references
        disc.finalize()
        return disc

    def __str__(self) -> str:
        s = ""
        s += "Discretization with ...\n"
        s += "{0:>10} nodes\n".format(len(self.nodes))

        for key, eles in self.elements.items():
            s += "{0:>10} {1} elements\n".format(len(eles), key)

        return s



import os
import unittest

import numpy as np
import sympy as sp
from lnmmeshio import (
    Element,
    Element1D,
    Element2D,
    Element3D,
    Hex8,
    Hex20,
    Hex27,
    Line2,
    Line3,
    Quad4,
    Quad8,
    Quad9,
    Tet4,
    Tet10,
    Tri3,
    Tri6,
    Vertex,
)
from lnmmeshio.element.parse_element import parse as parse_ele
from lnmmeshio.node import Node
from lnmmeshio.nodeset import LineNodeset, PointNodeset, SurfaceNodeset, VolumeNodeset
from parameterized import parameterized

script_dir = os.path.dirname(os.path.realpath(__file__))


class TestEles(unittest.TestCase):
    def setUp(self):
        pass

    def __get_nodes(self, n):
        nodes = [Node(np.random.rand(3)) for i in range(n)]

        for i, n in enumerate(nodes):
            n.id = i

        return nodes

    def test_common(self):
        ele = Element(None, "X", self.__get_nodes(3))

        vnds = [VolumeNodeset(0), VolumeNodeset(1), VolumeNodeset(2), VolumeNodeset(3)]
        snds = [
            SurfaceNodeset(10),
            SurfaceNodeset(11),
            SurfaceNodeset(12),
            SurfaceNodeset(13),
        ]
        lnds = [LineNodeset(20), LineNodeset(21), LineNodeset(22), LineNodeset(23)]
        pnds = [PointNodeset(30), PointNodeset(31), PointNodeset(32), PointNodeset(33)]

        for i in [0, 1, 2, 3]:
            vnds[i].add_node(ele.nodes[0])
        for i in [0, 1, 2]:
            vnds[i].add_node(ele.nodes[1])
        for i in [0, 1]:
            vnds[i].add_node(ele.nodes[2])

        for i in [0, 1, 2, 3]:
            snds[i].add_node(ele.nodes[0])
        for i in [0, 1, 2]:
            snds[i].add_node(ele.nodes[1])
        for i in [0, 1]:
            snds[i].add_node(ele.nodes[2])

        for i in [0, 1, 2, 3]:
            lnds[i].add_node(ele.nodes[0])
        for i in [0, 1, 2]:
            lnds[i].add_node(ele.nodes[1])
        for i in [0, 1]:
            lnds[i].add_node(ele.nodes[2])

        for i in [0, 1, 2, 3]:
            pnds[i].add_node(ele.nodes[0])
        for i in [0, 1, 2]:
            pnds[i].add_node(ele.nodes[1])
        for i in [0, 1]:
            pnds[i].add_node(ele.nodes[2])

        for ns in vnds:
            for n in ns:
                n.volumenodesets.append(ns)
        for ns in snds:
            for n in ns:
                n.surfacenodesets.append(ns)
        for ns in lnds:
            for n in ns:
                n.linenodesets.append(ns)
        for ns in pnds:
            for n in ns:
                n.pointnodesets.append(ns)

        self.assertListEqual(sorted([ns.id for ns in ele.get_dvols()]), [0, 1])
        self.assertListEqual(sorted([ns.id for ns in ele.get_dsurfs()]), [10, 11])
        self.assertListEqual(sorted([ns.id for ns in ele.get_dlines()]), [20, 21])
        self.assertListEqual(sorted([ns.id for ns in ele.get_dpoints()]), [30, 31])

        for i, n in enumerate(ele.nodes):
            n.id = i

        self.assertListEqual(list(ele.get_node_ids()), list(range(0, 3)))

    def __test_shape_functions_sum(self, ele):
        # check whether the sum of all shape functions is 1 everywhere
        if issubclass(ele, Element1D):
            xi = sp.symbols("xi_1")
        elif issubclass(ele, Element2D):
            xi = np.array([sp.symbols("xi_1"), sp.symbols("xi_2")])
        elif issubclass(ele, Element3D):
            xi = np.array([sp.symbols("xi_1"), sp.symbols("xi_2"), sp.symbols("xi_3")])
        else:
            raise NotImplementedError("Unknown number of dimension: {0}".format(ele))
        shapefcns = ele.shape_fcns(xi)

        self.assertAlmostEqual(float(sp.simplify(sum(shapefcns))), 1.0)

    @staticmethod
    def _subs(expr, x, y):
        if not hasattr(expr, "shape"):
            myexpr = expr
            for xi, yi in zip(x, y):
                myexpr = myexpr.subs(xi, yi)

            return myexpr

        out = np.zeros(expr.shape, dtype=expr.dtype)

        for index in np.ndindex(expr.shape):
            out[index] = expr[index]
            for xi, yi in zip(x, y):
                out[index] = out[index].subs(xi, yi)

        return out

    def __test_nodal_reference_coordinates(self, ele):
        # check whether all shape functions are 0 except the one of the reference coordinates
        if issubclass(ele, Element1D):
            xi = sp.symbols("xi_1")
        elif issubclass(ele, Element2D):
            xi = np.array([sp.symbols("xi_1"), sp.symbols("xi_2")])
        elif issubclass(ele, Element3D):
            xi = np.array([sp.symbols("xi_1"), sp.symbols("xi_2"), sp.symbols("xi_3")])
        else:
            raise NotImplementedError("Unknown number of dimension: {0}".format(ele))

        shapefcns = ele.shape_fcns(xi)
        for i, ref_coords in enumerate(ele.nodal_reference_coordinates()):
            for j, shapefcn in enumerate(shapefcns):
                self.assertAlmostEqual(
                    float(sp.simplify(self._subs(shapefcn, xi, ref_coords))),
                    1 if i == j else 0,
                )

    def __test_point_in_ref(self, cls):
        self.assertTrue(cls.is_in_ref(np.array([0.1, 0.1, 0.1])))
        self.assertTrue(cls.is_in_ref(np.array([1.0, 0.0, 0.0])))
        self.assertTrue(cls.is_in_ref(np.array([1.0, 0.0, 0.0]), include_boundary=True))
        self.assertFalse(
            cls.is_in_ref(np.array([1.0, 0.0, 0.0]), include_boundary=False)
        )
        self.assertFalse(cls.is_in_ref(np.array([-100, -100, -100])))

    def __test_ele(
        self, shape, cls, nnodes, faces, facetype, edges, edgetype, test_get_xi=False
    ):
        ele = parse_ele(
            "1 ELENAME {0} {1} MAT 11 KINEM nonlinear TYPE Std".format(
                shape, " ".join([str(i) for i in range(1, nnodes + 1)])
            ),
            self.__get_nodes(nnodes),
        )

        self.assertEqual(int(ele.options["MAT"]), 11)

        self.assertIsInstance(ele, cls)

        self.assertEqual(ele.shape, cls.ShapeName)
        self.assertEqual(ele.shape, shape)

        self.assertEqual(ele.get_num_nodes(), len(ele.nodes))
        self.assertEqual(ele.get_num_nodes(), nnodes)

        f = ele.get_faces()
        self.assertEqual(len(f), len(faces))
        for i, face in enumerate(f):
            self.assertEqual(face.shape, facetype.ShapeName)

            # check if faces and nodes are ordered correctly
            for j, n in enumerate(face.get_nodes()):
                self.assertEqual(n.id, faces[i][j])

        # test edges
        e = ele.get_edges()
        self.assertEqual(len(e), len(edges))
        for i, edge in enumerate(e):
            self.assertEqual(edge.shape, edgetype.ShapeName)

            # check if faces and nodes are ordered correctly
            for j, n in enumerate(edge.get_nodes()):
                self.assertEqual(
                    n.id, edges[i][j], "Error at edge {0} (node {1})".format(i, j)
                )

        if test_get_xi:
            coords = np.transpose(np.array([n.coords for n in ele.nodes]))
            for _ in range(100):
                xi = np.random.rand((3))

                # transform x to xi
                x = np.dot(coords, ele.shape_fcns(xi))

                # transform x to xi
                xi2 = ele.get_xi(x)

                # compare
                for a, b in zip(xi, xi2):
                    self.assertAlmostEqual(a, b)

    def _test_ele_shpfcns(self, cls, xis):
        for i, xi in enumerate(xis):
            if not isinstance(xi, np.ndarray):
                xis[i] = np.array(xi).reshape((-1))

        ndim = xis[0].shape[0]
        xi_minmax = np.full((ndim, 2), np.nan)
        for i, xi in enumerate(xis):
            shpfcns = cls.shape_fcns(xi)

            cmp_vct = np.zeros((len(shpfcns)), dtype=float)
            cmp_vct[i] = 1.0

            self.assertListEqual(
                list(shpfcns),
                list(cmp_vct),
                "Shapefunctions differ at {0} / xi={1}".format(i, xi),
            )

            for k, xij in enumerate(xi):
                if np.isnan(xi_minmax[k][0]) or xi_minmax[k][0] > xij:
                    xi_minmax[k][0] = xij
                if np.isnan(xi_minmax[k][1]) or xi_minmax[k][1] < xij:
                    xi_minmax[k][1] = xij

        constant = np.nan
        for _ in range(1000):
            r = np.random.rand(ndim)

            cur_result = np.sum(cls.shape_fcns(r))

            if np.isnan(constant):
                constant = cur_result
            else:
                self.assertAlmostEqual(constant, cur_result)

    @parameterized.expand(
        [
            [Hex8],
            [Hex20],
            [Hex27],
            [Tet4],
            [Tet10],
            [Line2],
            [Line3],
            [Quad4],
            [Quad8],
            [Quad9],
            [Tri3],
            [Tri6],
        ]
    )
    def test_shape_fcns(self, ele):
        self.__test_shape_functions_sum(ele)

    @parameterized.expand(
        [
            [Hex8],
            [Tet4],
            [Tet10],
            [Line2],
            [Quad4],
            [Tri3],
            [Tri6],
        ]
    )
    def test_shape_fcns_derivs(self, ele):
        # check whether the sum of all shape functions is 1 everywhere
        if issubclass(ele, Element1D):
            xi = sp.symbols("xi_1")
            num_xis = 1
        elif issubclass(ele, Element2D):
            xi = np.array([sp.symbols("xi_1"), sp.symbols("xi_2")])
            num_xis = 2
        elif issubclass(ele, Element3D):
            xi = np.array([sp.symbols("xi_1"), sp.symbols("xi_2"), sp.symbols("xi_3")])
            num_xis = 3
        else:
            raise NotImplementedError("Unknown number of dimension: {0}".format(ele))
        shapefcns = ele.shape_fcns(xi)
        shapefcnsderivs = ele.shape_fcns_derivs(xi).reshape(num_xis, -1)
        for i in range(num_xis):
            # compute symbolic derivative
            for n in range(ele.get_num_nodes()):
                analytic_deriv = sp.diff(shapefcns[n], xi[i] if num_xis > 1 else xi)
                self.assertTrue(
                    sp.simplify(shapefcnsderivs[i, n] - analytic_deriv) == 0
                )

    @parameterized.expand(
        [
            [Hex8, 3],
            [Hex20, 3],
            [Hex27, 3],
            [Tet4, 3],
            [Tet10, 3],
            [Line2, 1],
            [Line3, 1],
            [Quad4, 2],
            [Quad8, 2],
            [Quad9, 2],
            [Tri3, 2],
            [Tri6, 2],
        ]
    )
    def test_space_dim(self, ele, dim):
        self.assertEqual(ele.get_space_dim(), dim)

    @parameterized.expand(
        [
            [Hex8],
            [Tet4],
            [Tet10],
        ]
    )
    def test_nodal_reference_coordinates(self, ele):
        self.__test_nodal_reference_coordinates(ele)

    @parameterized.expand(
        [
            [Hex8],
            [Hex20],
            [Hex27],
            [Tet4],
            [Tet10],
        ]
    )
    def test_is_in_ref(self, ele):
        self.__test_point_in_ref(ele)

    def test_hex8(self):
        FACES = [
            [0, 1, 2, 3],
            [0, 1, 5, 4],
            [1, 2, 6, 5],
            [2, 3, 7, 6],
            [3, 0, 4, 7],
            [4, 5, 6, 7],
        ]
        EDGES = [
            [0, 1],
            [1, 2],
            [2, 3],
            [3, 0],
            [0, 4],
            [1, 5],
            [2, 6],
            [3, 7],
            [4, 5],
            [5, 6],
            [6, 7],
            [7, 4],
        ]

        # test ele
        self.__test_ele("HEX8", Hex8, 8, FACES, Quad4, EDGES, Line2)
        self._test_ele_shpfcns(
            Hex8,
            [
                np.array([-1.0, -1.0, -1.0]),
                np.array([1.0, -1.0, -1.0]),
                np.array([1.0, 1.0, -1.0]),
                np.array([-1.0, 1.0, -1.0]),
                np.array([-1.0, -1.0, 1.0]),
                np.array([1.0, -1.0, 1.0]),
                np.array([1.0, 1.0, 1.0]),
                np.array([-1.0, 1.0, 1.0]),
            ],
        )

    def test_hex20(self):
        FACES = [
            [0, 1, 2, 3, 8, 9, 10, 11],
            [0, 1, 5, 4, 8, 13, 16, 12],
            [1, 2, 6, 5, 9, 14, 17, 13],
            [2, 3, 7, 6, 10, 15, 18, 14],
            [3, 0, 4, 7, 11, 12, 19, 15],
            [4, 5, 6, 7, 15, 17, 18, 19],
        ]
        EDGES = [
            [0, 1, 8],
            [1, 2, 9],
            [2, 3, 10],
            [3, 0, 11],
            [0, 4, 12],
            [1, 5, 13],
            [2, 6, 14],
            [3, 7, 15],
            [4, 5, 16],
            [5, 6, 17],
            [6, 7, 18],
            [7, 4, 19],
        ]

        # test ele
        self.__test_ele("HEX20", Hex20, 20, FACES, Quad8, EDGES, Line3)
        self._test_ele_shpfcns(
            Hex20,
            [
                np.array([-1.0, -1.0, -1.0]),
                np.array([1.0, -1.0, -1.0]),
                np.array([1.0, 1.0, -1.0]),
                np.array([-1.0, 1.0, -1.0]),
                np.array([-1.0, -1.0, 1.0]),
                np.array([1.0, -1.0, 1.0]),
                np.array([1.0, 1.0, 1.0]),
                np.array([-1.0, 1.0, 1.0]),
                np.array([0.0, -1.0, -1.0]),
                np.array([1.0, 0.0, -1.0]),
                np.array([0.0, 1.0, -1.0]),
                np.array([-1.0, 0.0, -1.0]),
                np.array([-1.0, -1.0, 0.0]),
                np.array([1.0, -1.0, 0.0]),
                np.array([1.0, 1.0, 0.0]),
                np.array([-1.0, 1.0, 0.0]),
                np.array([0.0, -1.0, 1.0]),
                np.array([1.0, 0.0, 1.0]),
                np.array([0.0, 1.0, 1.0]),
                np.array([-1.0, 0.0, 1.0]),
            ],
        )

    def test_hex27(self):
        FACES = [
            [0, 1, 2, 3, 8, 9, 10, 11, 20],
            [0, 1, 5, 4, 8, 13, 16, 12, 21],
            [1, 2, 6, 5, 9, 14, 17, 13, 22],
            [2, 3, 7, 6, 10, 15, 18, 14, 23],
            [3, 0, 4, 7, 11, 12, 19, 15, 24],
            [4, 5, 6, 7, 15, 17, 18, 19, 25],
        ]
        EDGES = [
            [0, 1, 8],
            [1, 2, 9],
            [2, 3, 10],
            [3, 0, 11],
            [0, 4, 12],
            [1, 5, 13],
            [2, 6, 14],
            [3, 7, 15],
            [4, 5, 16],
            [5, 6, 17],
            [6, 7, 18],
            [7, 4, 19],
        ]

        self.__test_ele("HEX27", Hex27, 27, FACES, Quad9, EDGES, Line3)
        self._test_ele_shpfcns(
            Hex27,
            [
                np.array([-1.0, -1.0, -1.0]),
                np.array([1.0, -1.0, -1.0]),
                np.array([1.0, 1.0, -1.0]),
                np.array([-1.0, 1.0, -1.0]),
                np.array([-1.0, -1.0, 1.0]),
                np.array([1.0, -1.0, 1.0]),
                np.array([1.0, 1.0, 1.0]),
                np.array([-1.0, 1.0, 1.0]),
                np.array([0.0, -1.0, -1.0]),
                np.array([1.0, 0.0, -1.0]),
                np.array([0.0, 1.0, -1.0]),
                np.array([-1.0, 0.0, -1.0]),
                np.array([-1.0, -1.0, 0.0]),
                np.array([1.0, -1.0, 0.0]),
                np.array([1.0, 1.0, 0.0]),
                np.array([-1.0, 1.0, 0.0]),
                np.array([0.0, -1.0, 1.0]),
                np.array([1.0, 0.0, 1.0]),
                np.array([0.0, 1.0, 1.0]),
                np.array([-1.0, 0.0, 1.0]),
                np.array([0.0, 0.0, -1.0]),
                np.array([0.0, -1.0, 0.0]),
                np.array([1.0, 0.0, 0.0]),
                np.array([0.0, 1.0, 0.0]),
                np.array([-1.0, 0.0, 0.0]),
                np.array([0.0, 0.0, 1.0]),
                np.array([0.0, 0.0, 0.0]),
            ],
        )

    def test_tet10(self):
        FACES = [
            [0, 1, 3, 4, 8, 7],
            [1, 2, 3, 5, 9, 8],
            [2, 0, 3, 6, 7, 9],
            [0, 2, 1, 6, 5, 4],
        ]
        EDGES = [[0, 1, 4], [1, 2, 5], [2, 0, 6], [0, 3, 7], [1, 3, 8], [2, 3, 9]]

        # parse
        self.__test_ele("TET10", Tet10, 10, FACES, Tri6, EDGES, Line3)
        self._test_ele_shpfcns(
            Tet10,
            [
                np.array([0.0, 0.0, 0.0]),
                np.array([1.0, 0.0, 0.0]),
                np.array([0.0, 1.0, 0.0]),
                np.array([0.0, 0.0, 1.0]),
                np.array([0.5, 0.0, 0.0]),
                np.array([0.5, 0.5, 0.0]),
                np.array([0.0, 0.5, 0.0]),
                np.array([0.0, 0.0, 0.5]),
                np.array([0.5, 0.0, 0.5]),
                np.array([0.0, 0.5, 0.5]),
            ],
        )

    def test_tet4(self):
        FACES = [[0, 1, 3], [1, 2, 3], [2, 0, 3], [0, 2, 1]]
        EDGES = [[0, 1], [1, 2], [2, 0], [0, 3], [1, 3], [2, 3]]

        # parse
        self.__test_ele("TET4", Tet4, 4, FACES, Tri3, EDGES, Line2, test_get_xi=True)
        self._test_ele_shpfcns(
            Tet4,
            [
                np.array([0.0, 0.0, 0.0]),
                np.array([1.0, 0.0, 0.0]),
                np.array([0.0, 1.0, 0.0]),
                np.array([0.0, 0.0, 1.0]),
            ],
        )
        # test project quantity
        ele = Tet4(
            "",
            [
                Node(np.array([0.1, 0.2, 0.3])),
                Node(np.array([1.2, 0.3, 0.4])),
                Node(np.array([0.1, 2.2, 0.3])),
                Node(np.array([0.1, 0.22, 1.2])),
            ],
        )

        noddata = np.array([[0.1, 0.2], [0.2, 0.3], [0.3, 0.4], [0.4, 0.5]])

        for i in range(4):
            n1 = ele.nodes[i]
            q1 = noddata[i]
            for j in range(4):
                n2 = ele.nodes[j]
                q2 = noddata[j]

                for xi in np.linspace(0, 1, 10):
                    pos = n1.coords + xi * (n2.coords - n1.coords)
                    qxi = q1 + xi * (q2 - q1)

                    qxi_proj = ele.project_quantity(pos, noddata)

                    for i in range(len(qxi)):
                        self.assertAlmostEqual(qxi_proj[i], qxi[i])

    def test_quad4(self):
        FACES = [[0, 1, 2, 3]]
        EDGES = [[0, 1], [1, 2], [2, 3], [3, 0]]

        # parse
        self.__test_ele("QUAD4", Quad4, 4, FACES, Quad4, EDGES, Line2)
        self._test_ele_shpfcns(
            Quad4,
            [
                np.array([-1.0, -1.0]),
                np.array([1.0, -1.0]),
                np.array([1.0, 1.0]),
                np.array([-1.0, 1.0]),
            ],
        )

    def test_quad8(self):
        FACES = [[0, 1, 2, 3, 4, 5, 6, 7, 8]]
        EDGES = [[0, 1, 4], [1, 2, 5], [2, 3, 6], [3, 0, 7]]

        self.__test_ele("QUAD8", Quad8, 8, FACES, Quad8, EDGES, Line3)
        self._test_ele_shpfcns(
            Quad8,
            [
                np.array([-1.0, -1.0]),
                np.array([1.0, -1.0]),
                np.array([1.0, 1.0]),
                np.array([-1.0, 1.0]),
                np.array([0.0, -1.0]),
                np.array([1.0, 0.0]),
                np.array([0.0, 1.0]),
                np.array([-1.0, 0.0]),
            ],
        )

    def test_quad9(self):
        FACES = [[0, 1, 2, 3, 4, 5, 6, 7, 8, 9]]
        EDGES = [[0, 1, 4], [1, 2, 5], [2, 3, 6], [3, 0, 7]]

        self.__test_ele("QUAD9", Quad9, 9, FACES, Quad9, EDGES, Line3)
        self._test_ele_shpfcns(
            Quad9,
            [
                np.array([-1.0, -1.0]),
                np.array([1.0, -1.0]),
                np.array([1.0, 1.0]),
                np.array([-1.0, 1.0]),
                np.array([0.0, -1.0]),
                np.array([1.0, 0.0]),
                np.array([0.0, 1.0]),
                np.array([-1.0, 0.0]),
                np.array([0.0, 0.0]),
            ],
        )

    def test_tri3(self):
        FACES = [[0, 1, 2]]
        EDGES = [[0, 1], [1, 2], [2, 0]]

        self.__test_ele("TRI3", Tri3, 3, FACES, Tri3, EDGES, Line2)
        self._test_ele_shpfcns(
            Tri3, [np.array([0.0, 0.0]), np.array([1.0, 0.0]), np.array([0.0, 1.0])]
        )

    def test_tri6(self):
        FACES = [[0, 1, 2, 3, 4, 5]]
        EDGES = [[0, 1, 3], [1, 2, 4], [2, 0, 5]]

        self.__test_ele("TRI6", Tri6, 6, FACES, Tri6, EDGES, Line3)
        self._test_ele_shpfcns(
            Tri6,
            [
                np.array([0.0, 0.0]),
                np.array([1.0, 0.0]),
                np.array([0.0, 1.0]),
                np.array([0.5, 0.0]),
                np.array([0.5, 0.5]),
                np.array([0.0, 0.5]),
            ],
        )

    def test_line2(self):
        EDGES = [[0, 1]]

        self.__test_ele("LINE2", Line2, 2, [], None, EDGES, Line2)
        self._test_ele_shpfcns(Line2, [-1.0, 1.0])

    def test_line3(self):
        EDGES = [[0, 1, 2]]

        self.__test_ele("LINE3", Line3, 3, [], None, EDGES, Line3)
        self._test_ele_shpfcns(Line3, [-1.0, 1.0, 0.0])

    def test_vertex1(self):
        self.__test_ele("VERTEX1", Vertex, 1, [], None, [], None)



from typing import List

import numpy as np

from ..node import Node
from .element import ElementHex
from .line3 import Line3
from .quad8 import Quad8


class Hex20(ElementHex):
    """
    Implementation of a HEX20 element
    """

    ShapeName: str = "HEX20"

    def __init__(self, el_type: str, nodes: List[Node]):
        """
        Base constructor of a Hex20 element
        """
        super(Hex20, self).__init__(el_type, Hex20.ShapeName, nodes)

        if len(nodes) != self.get_num_nodes():
            raise RuntimeError(
                "You tried to created a HEX20 element with {0} nodes".format(len(nodes))
            )

    @classmethod
    def get_num_nodes(cls) -> int:
        """
        Get number of nodes of a HEX20 element

        Returns:
            Number of nodes of a Hex20 element = 20
        """
        return 20

    def get_faces(self) -> List[Quad8]:
        """
        Returns a list of faces of the Hex20 element

        Returns:
            List of faces
        """
        face_node_ids = [
            [0, 1, 2, 3, 8, 9, 10, 11],
            [0, 1, 5, 4, 8, 13, 16, 12],
            [1, 2, 6, 5, 9, 14, 17, 13],
            [2, 3, 7, 6, 10, 15, 18, 14],
            [3, 0, 4, 7, 11, 12, 19, 15],
            [4, 5, 6, 7, 15, 17, 18, 19],
        ]

        return [Quad8(None, [self.nodes[i] for i in nodes]) for nodes in face_node_ids]

    def get_edges(self) -> List[Line3]:
        """
        Returns the list of all edges

        Returns:
            List of edges
        """
        edge_node_ids = [
            [0, 1, 8],
            [1, 2, 9],
            [2, 3, 10],
            [3, 0, 11],
            [0, 4, 12],
            [1, 5, 13],
            [2, 6, 14],
            [3, 7, 15],
            [4, 5, 16],
            [5, 6, 17],
            [6, 7, 18],
            [7, 4, 19],
        ]
        return [Line3(None, [self.nodes[i] for i in nodes]) for nodes in edge_node_ids]

    @staticmethod
    def shape_fcns(xi):
        """
        Returns the value of the shape functions at the local coordinate xi
        """

        r = xi[0]
        s = xi[1]
        t = xi[2]

        rp = 1.0 + r
        rm = 1.0 - r
        sp = 1.0 + s
        sm = 1.0 - s
        tp = 1.0 + t
        tm = 1.0 - t
        rrm = 1.0 - r * r
        ssm = 1.0 - s * s
        ttm = 1.0 - t * t

        return np.array(
            [
                0.125 * rm * sm * tm * (rm + sm + tm - 5.0),
                0.125 * rp * sm * tm * (rp + sm + tm - 5.0),
                0.125 * rp * sp * tm * (rp + sp + tm - 5.0),
                0.125 * rm * sp * tm * (rm + sp + tm - 5.0),
                0.125 * rm * sm * tp * (rm + sm + tp - 5.0),
                0.125 * rp * sm * tp * (rp + sm + tp - 5.0),
                0.125 * rp * sp * tp * (rp + sp + tp - 5.0),
                0.125 * rm * sp * tp * (rm + sp + tp - 5.0),
                0.25 * rrm * sm * tm,
                0.25 * rp * ssm * tm,
                0.25 * rrm * sp * tm,
                0.25 * rm * ssm * tm,
                0.25 * rm * sm * ttm,
                0.25 * rp * sm * ttm,
                0.25 * rp * sp * ttm,
                0.25 * rm * sp * ttm,
                0.25 * rrm * sm * tp,
                0.25 * rp * ssm * tp,
                0.25 * rrm * sp * tp,
                0.25 * rm * ssm * tp,
            ]
        )



import os
import unittest

import lnmmeshio
import numpy as np

script_dir = os.path.dirname(os.path.realpath(__file__))


class TestMeshio2Discretization(unittest.TestCase):
    def setUp(self):
        pass

    def test_read_mesh(self):
        # read mesh

        dis: lnmmeshio.Discretization = lnmmeshio.read(
            os.path.join(script_dir, "data", "dummy.mesh")
        )

        if not os.path.isdir(os.path.join(script_dir, "tmp")):
            os.makedirs(os.path.join(script_dir, "tmp"))

        lnmmeshio.write(
            os.path.join(script_dir, "tmp", "dummy_mesh.dat"), dis, file_format="dat"
        )

    def test_vice_versa(self):
        dis: lnmmeshio.Discretization = lnmmeshio.read(
            os.path.join(script_dir, "data", "dummy2.dat")
        )

        mesh = lnmmeshio.meshio_to_discretization.discretization2mesh(dis)
        dis2: lnmmeshio.Discretization = (
            lnmmeshio.meshio_to_discretization.mesh2Discretization(mesh)
        )

        dis.compute_ids(zero_based=True)
        dis2.compute_ids(zero_based=True)

        # compare whether both discretizations are the same
        self.assertEqual(len(dis.nodes), len(dis2.nodes))

        # compare each node
        for node1, node2 in zip(dis.nodes, dis2.nodes):
            self.assertAlmostEqual(np.linalg.norm(node1.coords - node2.coords), 0)

        # compare nodesets
        # self.assertEqual(len(dis.pointnodesets), len(dis2.pointnodesets))
        # self.assertEqual(len(dis.linenodesets), len(dis2.linenodesets))
        # self.assertEqual(len(dis.surfacenodesets), len(dis2.surfacenodesets))

        # for pns1, pns2 in zip(dis.pointnodesets, dis2.pointnodesets):
        #    self.assertListEqual([n.id for n in pns1], [n.id for n in pns2])

        # for lns1, pns2 in zip(dis.linenodesets, dis2.linenodesets):
        #    self.assertListEqual([n.id for n in lns1], [n.id for n in lns1])

        # for sns1, pns2 in zip(dis.surfacenodesets, dis2.surfacenodesets):
        #    self.assertListEqual([n.id for n in sns1], [n.id for n in sns1])

        self.assertEqual(
            dis.elements.get_num_structure(), dis2.elements.get_num_structure()
        )
        self.assertEqual(dis.elements.get_num_fluid(), dis2.elements.get_num_fluid())
        self.assertEqual(dis.elements.get_num_ale(), dis2.elements.get_num_ale())
        self.assertEqual(
            dis.elements.get_num_transport(), dis2.elements.get_num_transport()
        )
        self.assertEqual(dis.elements.get_num_thermo(), dis2.elements.get_num_thermo())

        if dis.elements.get_num_structure() > 0:
            for ele1, ele2 in zip(dis.elements.structure, dis2.elements.structure):
                # check, whether nodes are the same within the element
                self.assertEqual(ele1.shape, ele2.shape)
                self.assertEqual(len(ele1.nodes), len(ele2.nodes))

                for n1, n2 in zip(ele1.nodes, ele2.nodes):
                    self.assertEqual(n1.id, n2.id)

        if dis.elements.get_num_ale() > 0:
            for ele1, ele2 in zip(dis.elements.ale, dis2.elements.ale):
                # check, whether nodes are the same within the element
                self.assertEqual(ele1.shape, ele2.shape)
                self.assertEqual(len(ele1.nodes), len(ele2.nodes))

                for n1, n2 in zip(ele1.nodes, ele2.nodes):
                    self.assertEqual(n1.id, n2.id)

        if dis.elements.get_num_fluid() > 0:
            for ele1, ele2 in zip(dis.elements.fluid, dis2.elements.fluid):
                # check, whether nodes are the same within the element
                self.assertEqual(ele1.shape, ele2.shape)
                self.assertEqual(len(ele1.nodes), len(ele2.nodes))

                for n1, n2 in zip(ele1.nodes, ele2.nodes):
                    self.assertEqual(n1.id, n2.id)

        if dis.elements.get_num_transport() > 0:
            for ele1, ele2 in zip(dis.elements.transport, dis2.elements.transport):
                # check, whether nodes are the same within the element
                self.assertEqual(ele1.shape, ele2.shape)
                self.assertEqual(len(ele1.nodes), len(ele2.nodes))

                for n1, n2 in zip(ele1.nodes, ele2.nodes):
                    self.assertEqual(n1.id, n2.id)

        if dis.elements.get_num_thermo() > 0:
            for ele1, ele2 in zip(dis.elements.thermo, dis2.elements.thermo):
                # check, whether nodes are the same within the element
                self.assertEqual(ele1.shape, ele2.shape)
                self.assertEqual(len(ele1.nodes), len(ele2.nodes))

                for n1, n2 in zip(ele1.nodes, ele2.nodes):
                    self.assertEqual(n1.id, n2.id)

    def test_2D_mesh_element_data(self):
        dis = lnmmeshio.Discretization()

        dis.nodes = [
            lnmmeshio.Node(np.array([1.0, 0.0, 0.0])),
            lnmmeshio.Node(np.array([1.0, 0.0, 0.0])),
            lnmmeshio.Node(np.array([1.0, 0.0, 0.0])),
            lnmmeshio.Node(np.array([1.0, 0.0, 0.0])),
            lnmmeshio.Node(np.array([1.0, 0.0, 0.0])),
        ]

        dis.elements.structure = [
            lnmmeshio.Tri3(None, [dis.nodes[0], dis.nodes[1], dis.nodes[2]]),
            lnmmeshio.Tri3(None, [dis.nodes[0], dis.nodes[3], dis.nodes[4]]),
        ]

        dis.elements.structure[0].data["test"] = 1
        dis.elements.structure[1].data["test"] = 2

        mesh = lnmmeshio.meshio_to_discretization.discretization2mesh(dis)

        self.assertListEqual(list(mesh.cell_data["test"][0]), [1, 2])



import os
import unittest

import numpy as np
from lnmmeshio import Fiber

script_dir = os.path.dirname(os.path.realpath(__file__))


class TestFibers(unittest.TestCase):
    def setUp(self):
        pass

    def test_fibers_parse(self):
        # parse
        ele_def = "1 TYPE SHAPE 1 2 3 MAT 1 FIBER1 0.1 0.2 0.3 FIBER2 0.2 0.3 0.4 FIBER3 0.01 0.02 0.03 FIBER4 0.04 0.05 0.06 FIBER5 0.05 0.06 0.07 FIBER6 0.08 0.08 0.1 FIBER7 0.2 0.1 0.01 FIBER8 0.3 0.2 0.2 FIBER9 0.01 0.2 0.9 CIR 0.3 0.4 0.5 TAN 0.4 0.5 0.6 RAD 0.5 0.1 0.1 AXI 0.5 0.6 0.8\n"

        # parse fibers
        fibs = Fiber.parse_fibers(ele_def)

        self.assertEqual(len(fibs), 13)

        self.assertAlmostEqual(
            np.linalg.norm(fibs[Fiber.TypeFiber1].fiber - np.array([0.1, 0.2, 0.3])),
            0.0,
        )
        self.assertAlmostEqual(
            np.linalg.norm(fibs[Fiber.TypeFiber2].fiber - np.array([0.2, 0.3, 0.4])),
            0.0,
        )
        self.assertAlmostEqual(
            np.linalg.norm(fibs[Fiber.TypeFiber3].fiber - np.array([0.01, 0.02, 0.03])),
            0.0,
        )
        self.assertAlmostEqual(
            np.linalg.norm(fibs[Fiber.TypeFiber4].fiber - np.array([0.04, 0.05, 0.06])),
            0.0,
        )
        self.assertAlmostEqual(
            np.linalg.norm(fibs[Fiber.TypeFiber5].fiber - np.array([0.05, 0.06, 0.07])),
            0.0,
        )
        self.assertAlmostEqual(
            np.linalg.norm(fibs[Fiber.TypeFiber6].fiber - np.array([0.08, 0.08, 0.1])),
            0.0,
        )
        self.assertAlmostEqual(
            np.linalg.norm(fibs[Fiber.TypeFiber7].fiber - np.array([0.2, 0.1, 0.01])),
            0.0,
        )
        self.assertAlmostEqual(
            np.linalg.norm(fibs[Fiber.TypeFiber8].fiber - np.array([0.3, 0.2, 0.2])),
            0.0,
        )
        self.assertAlmostEqual(
            np.linalg.norm(fibs[Fiber.TypeFiber9].fiber - np.array([0.01, 0.2, 0.9])),
            0.0,
        )
        self.assertAlmostEqual(
            np.linalg.norm(fibs[Fiber.TypeCir].fiber - np.array([0.3, 0.4, 0.5])), 0.0
        )
        self.assertAlmostEqual(
            np.linalg.norm(fibs[Fiber.TypeTan].fiber - np.array([0.4, 0.5, 0.6])), 0.0
        )
        self.assertAlmostEqual(
            np.linalg.norm(fibs[Fiber.TypeAxi].fiber - np.array([0.5, 0.6, 0.8])), 0.0
        )
        self.assertAlmostEqual(
            np.linalg.norm(fibs[Fiber.TypeRad].fiber - np.array([0.5, 0.1, 0.1])), 0.0
        )

        self.assertEqual(Fiber.get_fiber_type("FIBER1"), Fiber.TypeFiber1)
        self.assertEqual(Fiber.get_fiber_type("FIBER2"), Fiber.TypeFiber2)
        self.assertEqual(Fiber.get_fiber_type("FIBER3"), Fiber.TypeFiber3)
        self.assertEqual(Fiber.get_fiber_type("FIBER4"), Fiber.TypeFiber4)
        self.assertEqual(Fiber.get_fiber_type("FIBER5"), Fiber.TypeFiber5)
        self.assertEqual(Fiber.get_fiber_type("FIBER6"), Fiber.TypeFiber6)
        self.assertEqual(Fiber.get_fiber_type("FIBER7"), Fiber.TypeFiber7)
        self.assertEqual(Fiber.get_fiber_type("FIBER8"), Fiber.TypeFiber8)
        self.assertEqual(Fiber.get_fiber_type("FIBER9"), Fiber.TypeFiber9)
        self.assertEqual(Fiber.get_fiber_type("CIR"), Fiber.TypeCir)
        self.assertEqual(Fiber.get_fiber_type("TAN"), Fiber.TypeTan)
        self.assertEqual(Fiber.get_fiber_type("AXI"), Fiber.TypeAxi)
        self.assertEqual(Fiber.get_fiber_type("RAD"), Fiber.TypeRad)



from typing import List

import numpy as np

from ..node import Node
from .element import ElementQuad
from .line3 import Line3


class Quad9(ElementQuad):
    """
    Implementation of a quad9 element
    """

    ShapeName: str = "QUAD9"

    def __init__(self, el_type: str, nodes: List[Node]):
        """
        Base constructor of a Quad9 element
        """
        super(Quad9, self).__init__(el_type, Quad9.ShapeName, nodes)

        if len(nodes) != self.get_num_nodes():
            raise RuntimeError(
                "You tried to created a QUAD9 element with {0} nodes".format(len(nodes))
            )

    @classmethod
    def get_num_nodes(cls) -> int:
        """
        Get number of nodes of a QUAD9 element

        Returns:
            Number of nodes of a QUAD9 element = 9
        """
        return 9

    def get_faces(self) -> List["Quad9"]:
        """
        The face of a quad9 element is itself

        Returns:
            List of faces
        """
        return [self]

    def get_edges(self) -> List[Line3]:
        """
        Returns the list of all edges

        Returns:
            List of edges
        """
        return [
            Line3(None, [self.nodes[0], self.nodes[1], self.nodes[4]]),
            Line3(None, [self.nodes[1], self.nodes[2], self.nodes[5]]),
            Line3(None, [self.nodes[2], self.nodes[3], self.nodes[6]]),
            Line3(None, [self.nodes[3], self.nodes[0], self.nodes[7]]),
        ]

    @staticmethod
    def shape_fcns(xi):
        """
        Returns the value of the shape functions at the local coordinate xi
        """

        return np.array(
            [
                Line3.shape_fcns(xi[0])[0] * Line3.shape_fcns(xi[1])[0],
                Line3.shape_fcns(xi[0])[1] * Line3.shape_fcns(xi[1])[0],
                Line3.shape_fcns(xi[0])[1] * Line3.shape_fcns(xi[1])[1],
                Line3.shape_fcns(xi[0])[0] * Line3.shape_fcns(xi[1])[1],
                Line3.shape_fcns(xi[0])[2] * Line3.shape_fcns(xi[1])[0],
                Line3.shape_fcns(xi[0])[1] * Line3.shape_fcns(xi[1])[2],
                Line3.shape_fcns(xi[0])[2] * Line3.shape_fcns(xi[1])[1],
                Line3.shape_fcns(xi[0])[0] * Line3.shape_fcns(xi[1])[2],
                Line3.shape_fcns(xi[0])[2] * Line3.shape_fcns(xi[1])[2],
            ]
        )



from typing import IO, Iterator, List, Optional, Set

from loguru import logger
from tqdm import tqdm

from .ioutils import read_option_item, write_title
from .node import Node


class Nodeset:
    def __init__(self, id: int, name: Optional[str] = None):
        self.id: Optional[int] = id
        self.nodes: Set[Node] = set()
        self.name: Optional[str] = name

    @staticmethod
    def get_typename_long() -> str:
        raise NotImplementedError("Need to implement get_typename_long()")

    @staticmethod
    def get_typename_short() -> str:
        raise NotImplementedError("Need to implement get_typename_short()")

    def reset(self) -> None:
        self.id = None

    def add_node(self, node: Node) -> None:
        if node not in self.nodes:
            self.nodes.add(node)

    def add_nodes(self, nodes: List[Node]) -> None:
        for n in nodes:
            self.nodes.add(n)

    def __len__(self) -> int:
        return len(self.nodes)

    def __iter__(self) -> Iterator:
        return iter(self.nodes)

    def __getitem__(self, x: int) -> Node:
        return list(self.nodes)[x]

    def get_lines(self) -> List[str]:
        lines = []
        for n in self:
            lines.append(
                "NODE {0} D{1} {2}".format(
                    n.id, self.get_typename_long().upper(), self.id
                )
            )

        return lines

    def write(self, dest: IO) -> None:
        for l in self.get_lines():
            dest.write("{0}\n".format(l))

    @classmethod
    def read(cls, lines: List[str], nodes, out: bool = False) -> List["Nodeset"]:
        id2pos = {}
        nodesets = []

        next_number = 0
        for line in tqdm(lines, disable=not out, desc="dnode topology"):
            nodeid_str, _ = read_option_item(line, "NODE")
            if nodeid_str is None or nodeid_str == "":
                # this is not a node, probably a comment
                continue

            try:
                nodeid = int(nodeid_str)
            except ValueError:
                logger.warning(f"Could not read {nodeid_str} as int")
                continue
            dpoint = int(
                read_option_item(line, "D{0}".format(cls.get_typename_long()))[0]
            )

            if dpoint not in id2pos:
                id2pos[dpoint] = next_number
                next_number += 1
                nodesets.append(cls(dpoint))

            nodesets[id2pos[dpoint]].add_node(nodes[nodeid - 1])

        return nodesets

    @staticmethod
    def get_section() -> str:
        raise RuntimeError("This function should not be called")

    @classmethod
    def write_header(cls, dest: IO) -> None:
        write_title(dest, cls.get_section())


class PointNodeset(Nodeset):
    def __init__(self, id, name=None):
        super(PointNodeset, self).__init__(id, name=name)

    @staticmethod
    def get_typename_long() -> str:
        return "NODE"

    @staticmethod
    def get_typename_short() -> str:
        return "NODE"

    @staticmethod
    def get_section() -> str:
        return "DNODE-NODE TOPOLOGY"

    @staticmethod
    def write_header(dest: IO) -> None:
        write_title(dest, PointNodeset.get_section())


class LineNodeset(Nodeset):
    def __init__(self, id, name=None):
        super(LineNodeset, self).__init__(id, name=name)

    @staticmethod
    def get_typename_long():
        return "LINE"

    @staticmethod
    def get_typename_short():
        return "LINE"

    @staticmethod
    def get_section():
        return "DLINE-NODE TOPOLOGY"

    @staticmethod
    def write_header(dest):
        write_title(dest, LineNodeset.get_section())


class SurfaceNodeset(Nodeset):
    def __init__(self, id, name=None):
        super(SurfaceNodeset, self).__init__(id, name=name)

    @staticmethod
    def get_typename_long():
        return "SURFACE"

    @staticmethod
    def get_typename_short():
        return "SURF"

    @staticmethod
    def get_section():
        return "DSURF-NODE TOPOLOGY"

    @staticmethod
    def write_header(dest):
        write_title(dest, SurfaceNodeset.get_section())


class VolumeNodeset(Nodeset):
    def __init__(self, id, name=None):
        super(VolumeNodeset, self).__init__(id, name=name)

    @staticmethod
    def get_typename_long():
        return "VOL"

    @staticmethod
    def get_typename_short():
        return "VOL"

    @staticmethod
    def get_section():
        return "DVOL-NODE TOPOLOGY"

    @staticmethod
    def write_header(dest):
        write_title(dest, VolumeNodeset.get_section())


class NodesetBuilder:
    def __init__(self, nstype):
        self.nstype = nstype
        self.nodesets = []
        self.id2pos = {}

    def add(self, node, id):
        if id not in self.id2pos:
            self.id2pos[id] = len(self.nodesets)
            self.nodesets.append(self.nstype(id))

        self.nodesets[self.id2pos[id]].add_node(node)

    def get_unused_id(self):
        nsid = 1
        while nsid in self.id2pos.keys():
            nsid += 1

        return nsid

    def finalize(self):
        # return nodeset sorted by its id to preserve the id
        return [self.nodesets[self.id2pos[i]] for i in sorted(list(self.id2pos))]


class PointNodesetBuilder(NodesetBuilder):
    def __init__(self):
        super(PointNodesetBuilder, self).__init__(PointNodeset)


class LineNodesetBuilder(NodesetBuilder):
    def __init__(self):
        super(LineNodesetBuilder, self).__init__(LineNodeset)


class SurfaceNodesetBuilder(NodesetBuilder):
    def __init__(self):
        super(SurfaceNodesetBuilder, self).__init__(SurfaceNodeset)


class VolumeNodesetBuilder(NodesetBuilder):
    def __init__(self):
        super(VolumeNodesetBuilder, self).__init__(VolumeNodeset)



import io
import re
import struct
from typing import IO, Any, Callable, Dict, Iterable, Iterator, List, Optional, Tuple

import numpy as np
from tqdm import tqdm


def read_dat_sections(origin):
    """
    Reads a dat file format and returns a dictionary containing of the sections with their lines

    Args:
        origin: File hander to read from

    Returns:
        dict: Dictionary with section names as key and lines as value
    """
    content = {}

    re_title = re.compile(r"^-{3,}(.*)")

    current_section = ""
    content[current_section] = []
    for line in tqdm(origin, desc="Read input"):
        line_no_comment = line.split("//", 1)[0].strip()
        match_title = re_title.match(line_no_comment)
        if match_title:
            # this is a section title
            current_section = match_title.group(1)
            if current_section in content:
                raise ValueError("{0} is dublicate!".format(current_section))

            content[current_section] = []
        else:
            content[current_section].append(line)

    return content


def read_option_item(line: str, option: str) -> Tuple[str, Tuple[int, int]]:
    values, span = read_option_items(line, option, 1)

    return values[0], span


def read_option_items(
    line: str, option: str, num: int = 1
) -> Tuple[List[str], Tuple[int, int]]:
    regex = re.compile(
        r"(^| ){0}{1}($|\s)".format(re.escape(option), num * "[ ]+([\\S]+)")
    )

    # split comment
    line = line.split("//", 1)[0]

    # read option
    match = regex.search(line)

    if not match:
        raise RuntimeError(f"Option {option} not found")

    return [match.group(i) for i in range(2, num + 2)], match.span(0)


def read_ints(line: str, option: str, num: int) -> np.ndarray:
    str_items, _ = read_option_items(line, option, num)

    int_items = [int(i) for i in str_items]
    return np.array(int_items)


def read_int(line: str, option: str) -> int:
    return read_ints(line, option, 1)[0]


def read_floats(line: str, option: str, num: int) -> np.ndarray:
    str_items, _ = read_option_items(line, option, num)
    return np.array([float(i) for i in str_items])


def read_float(line: str, option: str) -> int:
    return read_floats(line, option, 1)[0]


def read_key_values(
    line: str, num_items: Callable[[str], int]
) -> Iterable[Tuple[str, List[str]]]:
    regex = re.compile(r"^[ ]*(\S+)\s*")

    # split comment
    line = line.split("//", 1)[0]

    while True:
        match = regex.search(line)
        if not match:
            return

        line = line[match.span(0)[1] :]
        key = match.group(1)
        values: List[str] = []

        for _ in range(num_items(key)):
            match = regex.search(line)

            if not match:
                raise RuntimeError(f"Error reading option {key}")

            values.append(match.group(1))
            line = line[match.span(0)[1] :]

        yield key, values


# def read_next_key(line: str) -> Tuple[str, str]:
#     regex = re.compile(r"^[ ]*(\S+)\s*")

#     # split comment
#     line = line.split("//", 1)[0]

#     # read option
#     match = regex.search(line)

#     if not match:
#         return None, None

#     # shorten line by parsed option
#     line = line[match.span(0)[1] :]
#     return line, match.group(1)


# def read_next_value(line, num: int = 1):
#     regex = re.compile(
#         "^[ ]*{0}\\s*".format("[ ]+".join(["([\\S]+)" for i in range(0, num)]))
#     )

#     # split comment
#     line = line.split("//", 1)[0]

#     # read option
#     match = regex.search(line)

#     if not match:
#         return None, None, None

#     # shorten line by parsed option
#     line = line[match.span(0)[1] :]

#     return line, [match.group(i) for i in range(1, num + 1)]


# def read_next_option(line: str, num: int = 1):
#     regex = re.compile("^[ ]*(\\S+){0}\\s*".format(num * "[ ]+([\\S]+)"))

#     # split comment
#     line = line.split("//", 1)[0]

#     # read option
#     match = regex.search(line)

#     if not match:
#         return None, None, None

#     # shorten line by parsed option
#     line = line[match.span(0)[1] :]
#     match.group(1)
#     match.group(2)

#     return line, match.group(1), [match.group(i) for i in range(2, num + 2)]


def text_fill(
    text: str, length: int, chr: str = " ", minimum: int = 1, fill_left: bool = False
) -> str:
    fill_str: str = chr * max(minimum, length - len(text))

    if fill_left:
        return "{1}{0}".format(text, fill_str)
    else:
        return "{0}{1}".format(text, fill_str)


def line_title(title: str) -> str:
    return text_fill(title, 73, chr="-", minimum=3, fill_left=True)


def write_title(dest: IO, title: str, newline: bool = True) -> None:
    dest.write("{0}{1}".format(line_title(title), "\n" if newline else ""))


def line_option(key: str, value: Any, comment: Optional[str] = None) -> str:
    if comment is not None:
        return "{0}{1}// {2}".format(
            text_fill(key, 32, chr=" "), text_fill(str(value), 32, chr=" "), comment
        )
    else:
        return "{0}{1}".format(text_fill(key, 32, chr=" "), str(value))


def write_option(
    dest: IO, key: str, value: Any, comment: Optional[str] = None, newline: bool = True
) -> None:
    dest.write(
        "{0}{1}".format(line_option(key, value, comment), "\n" if newline else "")
    )


def line_option_list(list: Dict[str, np.ndarray]) -> str:
    """
    Return a line with the options and values asa defined in the dictionary

    Hint: This function accepts an ordered dict. Unfortunately, Python 3.6 - 3.7.2, we cannot set this as the type.
    """
    line = io.StringIO()
    first_entry: bool = True
    for key, value in list.items():
        if not first_entry:
            line.write(" ")
        first_entry = False

        if hasattr(value, "__iter__") and not isinstance(value, str):
            val_str = " ".join([str(i) for i in value])
        else:
            val_str = str(value)

        line.write("{0} {1}".format(key, val_str))

    return line.getvalue()


def write_option_list(
    dest: IO, options: Dict[str, np.ndarray], newline: bool = True
) -> None:
    dest.write("{0}{1}".format(line_option_list(options), "\n" if newline else ""))


def line_comment(comment: str) -> str:
    return "// {0}".format(comment)


def write_comment(dest: IO, comment: str, newline: bool = True) -> None:
    dest.write("{0}{1}".format(line_comment(comment), "\n" if newline else ""))


def ens_write_floats(file_handle: IO, arr: np.ndarray, binary: bool = True) -> None:
    """reads array of length floats"""
    if binary:
        np.ravel(arr, "F").astype("<f").tofile(file_handle)
    else:
        for f in np.ravel(arr, "F").astype("<f"):
            ens_write_float(file_handle, f, binary)


def ens_write_ints(file_handle: IO, arr: np.ndarray, binary: bool = True) -> None:
    if binary:
        np.ravel(arr).astype("<i").tofile(file_handle)
    else:
        if len(arr.shape) == 2:
            for i in range(arr.shape[0]):
                for j in range(arr.shape[1]):
                    ens_write_int(
                        file_handle, int(arr[i, j]), binary=binary, newline=False
                    )
                ens_write_string(file_handle, "", False)
        else:
            for i in np.ravel(arr, "F").astype("<i"):
                ens_write_int(file_handle, i, binary=binary)


def ens_write_int(
    file_handle: IO, i: int, binary: bool = True, newline: bool = True
) -> None:
    if binary:
        file_handle.write(struct.pack("i", i))
    else:
        s = str(i)
        file_handle.write(
            " " * (10 - len(s)) + s + "{0}".format("\n" if newline else "")
        )


def ens_write_float(file_handle: IO, v: float, binary: bool = True) -> None:
    if binary:
        file_handle.write(struct.pack("f", v))
    else:
        if v >= 0:
            file_handle.write(" ")
        file_handle.write("{0:10.5e}\n".format(v))


def ens_write_string(file_handle: IO, s: str, binary: bool = True) -> None:
    ws = ""

    if binary:
        ws = s[:80]
        ws = s + "\0" * (80 - len(s))
        ws = ws.encode("ascii")
    else:
        ws = s + "\n"

    file_handle.write(ws)



from typing import List

import numpy as np

from ..node import Node
from .element import ElementHex
from .line3 import Line3
from .quad9 import Quad9


class Hex27(ElementHex):
    """
    Implementation of a HEX27 element
    """

    ShapeName: str = "HEX27"

    def __init__(self, el_type: str, nodes: List[Node]):
        """
        Base constructor of a Hex27 element
        """
        super(Hex27, self).__init__(el_type, Hex27.ShapeName, nodes)

        if len(nodes) != self.get_num_nodes():
            raise RuntimeError(
                "You tried to created a HEX27 element with {0} nodes".format(len(nodes))
            )

    @classmethod
    def get_num_nodes(cls) -> int:
        """
        Get number of nodes of a HEX27 element

        Returns:
            Number of nodes of a Hex27 element = 27
        """
        return 27

    def get_faces(self) -> List[Quad9]:
        """
        Returns a list of faces of the Hex27 element

        Returns:
            List of faces
        """
        face_node_ids = [
            [0, 1, 2, 3, 8, 9, 10, 11, 20],
            [0, 1, 5, 4, 8, 13, 16, 12, 21],
            [1, 2, 6, 5, 9, 14, 17, 13, 22],
            [2, 3, 7, 6, 10, 15, 18, 14, 23],
            [3, 0, 4, 7, 11, 12, 19, 15, 24],
            [4, 5, 6, 7, 15, 17, 18, 19, 25],
        ]

        return [Quad9(None, [self.nodes[i] for i in nodes]) for nodes in face_node_ids]

    def get_edges(self) -> List[Line3]:
        """
        Returns the list of all edges

        Returns:
            List of edges
        """
        edge_node_ids = [
            [0, 1, 8],
            [1, 2, 9],
            [2, 3, 10],
            [3, 0, 11],
            [0, 4, 12],
            [1, 5, 13],
            [2, 6, 14],
            [3, 7, 15],
            [4, 5, 16],
            [5, 6, 17],
            [6, 7, 18],
            [7, 4, 19],
        ]
        return [Line3(None, [self.nodes[i] for i in nodes]) for nodes in edge_node_ids]

    @staticmethod
    def shape_fcns(xi):
        """
        Returns the value of the shape functions at the local coordinate xi
        """
        return np.array(
            [
                Line3.shape_fcns(xi[0])[0]
                * Line3.shape_fcns(xi[1])[0]
                * Line3.shape_fcns(xi[2])[0],
                Line3.shape_fcns(xi[0])[1]
                * Line3.shape_fcns(xi[1])[0]
                * Line3.shape_fcns(xi[2])[0],
                Line3.shape_fcns(xi[0])[1]
                * Line3.shape_fcns(xi[1])[1]
                * Line3.shape_fcns(xi[2])[0],
                Line3.shape_fcns(xi[0])[0]
                * Line3.shape_fcns(xi[1])[1]
                * Line3.shape_fcns(xi[2])[0],
                Line3.shape_fcns(xi[0])[0]
                * Line3.shape_fcns(xi[1])[0]
                * Line3.shape_fcns(xi[2])[1],
                Line3.shape_fcns(xi[0])[1]
                * Line3.shape_fcns(xi[1])[0]
                * Line3.shape_fcns(xi[2])[1],
                Line3.shape_fcns(xi[0])[1]
                * Line3.shape_fcns(xi[1])[1]
                * Line3.shape_fcns(xi[2])[1],
                Line3.shape_fcns(xi[0])[0]
                * Line3.shape_fcns(xi[1])[1]
                * Line3.shape_fcns(xi[2])[1],
                Line3.shape_fcns(xi[0])[2]
                * Line3.shape_fcns(xi[1])[0]
                * Line3.shape_fcns(xi[2])[0],
                Line3.shape_fcns(xi[0])[1]
                * Line3.shape_fcns(xi[1])[2]
                * Line3.shape_fcns(xi[2])[0],
                Line3.shape_fcns(xi[0])[2]
                * Line3.shape_fcns(xi[1])[1]
                * Line3.shape_fcns(xi[2])[0],
                Line3.shape_fcns(xi[0])[0]
                * Line3.shape_fcns(xi[1])[2]
                * Line3.shape_fcns(xi[2])[0],
                Line3.shape_fcns(xi[0])[0]
                * Line3.shape_fcns(xi[1])[0]
                * Line3.shape_fcns(xi[2])[2],
                Line3.shape_fcns(xi[0])[1]
                * Line3.shape_fcns(xi[1])[0]
                * Line3.shape_fcns(xi[2])[2],
                Line3.shape_fcns(xi[0])[1]
                * Line3.shape_fcns(xi[1])[1]
                * Line3.shape_fcns(xi[2])[2],
                Line3.shape_fcns(xi[0])[0]
                * Line3.shape_fcns(xi[1])[1]
                * Line3.shape_fcns(xi[2])[2],
                Line3.shape_fcns(xi[0])[2]
                * Line3.shape_fcns(xi[1])[0]
                * Line3.shape_fcns(xi[2])[1],
                Line3.shape_fcns(xi[0])[1]
                * Line3.shape_fcns(xi[1])[2]
                * Line3.shape_fcns(xi[2])[1],
                Line3.shape_fcns(xi[0])[2]
                * Line3.shape_fcns(xi[1])[1]
                * Line3.shape_fcns(xi[2])[1],
                Line3.shape_fcns(xi[0])[0]
                * Line3.shape_fcns(xi[1])[2]
                * Line3.shape_fcns(xi[2])[1],
                Line3.shape_fcns(xi[0])[2]
                * Line3.shape_fcns(xi[1])[2]
                * Line3.shape_fcns(xi[2])[0],
                Line3.shape_fcns(xi[0])[2]
                * Line3.shape_fcns(xi[1])[0]
                * Line3.shape_fcns(xi[2])[2],
                Line3.shape_fcns(xi[0])[1]
                * Line3.shape_fcns(xi[1])[2]
                * Line3.shape_fcns(xi[2])[2],
                Line3.shape_fcns(xi[0])[2]
                * Line3.shape_fcns(xi[1])[1]
                * Line3.shape_fcns(xi[2])[2],
                Line3.shape_fcns(xi[0])[0]
                * Line3.shape_fcns(xi[1])[2]
                * Line3.shape_fcns(xi[2])[2],
                Line3.shape_fcns(xi[0])[2]
                * Line3.shape_fcns(xi[1])[2]
                * Line3.shape_fcns(xi[2])[1],
                Line3.shape_fcns(xi[0])[2]
                * Line3.shape_fcns(xi[1])[2]
                * Line3.shape_fcns(xi[2])[2],
            ]
        )



from typing import List

from ..node import Node
from .element import Element0D


class Vertex(Element0D):
    """
    Implementation of a Vertex element
    """

    ShapeName: str = "VERTEX1"

    def __init__(self, el_type: str, nodes: List[Node]):
        """
        Base constructor of a Vertex element
        """
        super(Vertex, self).__init__(el_type, Vertex.ShapeName, nodes)

        if len(nodes) != self.get_num_nodes():
            raise RuntimeError(
                "You tried to created a Vertex element with {0} nodes".format(
                    len(nodes)
                )
            )

    @classmethod
    def get_num_nodes(cls) -> int:
        """
        Get number of nodes of a VERTEX element

        Returns:
            Number of nodes of a VERTEX element = 1
        """
        return 1

    def get_faces(self) -> List:
        """
        A VERTEX element does not contain any faces

        Returns:
            List of faces
        """
        return []

    def get_edges(self) -> List:
        """
        Returns the list of all edges

        Returns:
            List of edges
        """
        return []



import math
from typing import List

import numpy as np

from ..node import Node
from .element import ElementQuad
from .line2 import Line2


class Quad4(ElementQuad):
    """
    Implementation of a quad4 element
    """

    ShapeName: str = "QUAD4"

    def __init__(self, el_type: str, nodes: List[Node]):
        """
        Base constructor of a Quad4 element
        """
        super(Quad4, self).__init__(el_type, Quad4.ShapeName, nodes)

        if len(nodes) != self.get_num_nodes():
            raise RuntimeError(
                "You tried to created a QUAD4 element with {0} nodes".format(len(nodes))
            )

    @classmethod
    def get_num_nodes(cls) -> int:
        """
        Get number of nodes of a QUAD4 element

        Returns:
            Number of nodes of a QUAD4 element = 4
        """
        return 4

    def get_faces(self) -> List["Quad4"]:
        """
        The face of a quad4 element is itself

        Returns:
            List of faces
        """
        return [self]

    def get_edges(self) -> List[Line2]:
        """
        Returns the list of all edges

        Returns:
            List of edges
        """
        return [
            Line2(None, [self.nodes[0], self.nodes[1]]),
            Line2(None, [self.nodes[1], self.nodes[2]]),
            Line2(None, [self.nodes[2], self.nodes[3]]),
            Line2(None, [self.nodes[3], self.nodes[0]]),
        ]

    @staticmethod
    def shape_fcns(xi):
        """
        Returns the value of the shape functions at the local coordinate xi
        """

        return np.array(
            [
                Line2.shape_fcns(xi[0])[0] * Line2.shape_fcns(xi[1])[0],
                Line2.shape_fcns(xi[0])[1] * Line2.shape_fcns(xi[1])[0],
                Line2.shape_fcns(xi[0])[1] * Line2.shape_fcns(xi[1])[1],
                Line2.shape_fcns(xi[0])[0] * Line2.shape_fcns(xi[1])[1],
            ]
        )

    @staticmethod
    def shape_fcns_derivs(xi):
        """
        Returns the value of the derivatives of the shape functions with respect to the local coordinates at the local coordinate xi

        +-                                  -+
        |  dN_1 / dxi_1   dN_2 / dxi_1, ...  |
        |  dN_1 / dxi_2   dN_2 / dxi_2, ...  |
        +-                                  -+
        """

        return np.array(
            [
                [
                    Line2.shape_fcns_derivs(xi[0])[0] * Line2.shape_fcns(xi[1])[0],
                    Line2.shape_fcns_derivs(xi[0])[1] * Line2.shape_fcns(xi[1])[0],
                    Line2.shape_fcns_derivs(xi[0])[1] * Line2.shape_fcns(xi[1])[1],
                    Line2.shape_fcns_derivs(xi[0])[0] * Line2.shape_fcns(xi[1])[1],
                ],
                [
                    Line2.shape_fcns(xi[0])[0] * Line2.shape_fcns_derivs(xi[1])[0],
                    Line2.shape_fcns(xi[0])[1] * Line2.shape_fcns_derivs(xi[1])[0],
                    Line2.shape_fcns(xi[0])[1] * Line2.shape_fcns_derivs(xi[1])[1],
                    Line2.shape_fcns(xi[0])[0] * Line2.shape_fcns_derivs(xi[1])[1],
                ],
            ]
        )

    def integrate_xi(self, integrand, numgp) -> np.ndarray:
        """
        Integrates the integrand over the element

        Args:
            integrand: integrand as a function of xi
            numgp: Number of integration points to be used
        """
        result = np.array(0.0 * integrand(np.zeros((3))))

        intpoints: np.ndarray = ElementQuad.int_points(numgp)
        intweights: np.ndarray = ElementQuad.int_weights(numgp)

        coords = np.array([n.coords for n in self.nodes])

        for xi, w in zip(intpoints, intweights):
            J = np.matmul(Quad4.shape_fcns_derivs(xi), coords)

            # covariant metric tensor
            G = np.matmul(J, J.T)
            detA = math.sqrt(np.linalg.det(G))
            result += w * detA * integrand(xi)

        return result

    def get_normal(self, xi):
        """
        This is just an approximation assuming that the surface is plane
        """
        u = self.nodes[1].coords - self.nodes[0].coords
        v = self.nodes[2].coords - self.nodes[0].coords
        n = np.cross(u, v)
        return n / np.linalg.norm(n)



from typing import List

from .. import node
from .element import (
    Element,
    Element1D,
    Element2D,
    Element3D,
    ElementHex,
    ElementTet,
    ElementTri,
)
from .parse_element import create_element


def factory(
    shape: str, nodes: List[node.Node], type: str = None, throw_if_unknown: bool = True
) -> Element:
    return create_element(type, shape, nodes, throw_if_unknown=throw_if_unknown)



from typing import List

import numpy as np

from ..node import Node
from .element import ElementHex
from .line2 import Line2
from .quad4 import Quad4


class Hex8(ElementHex):
    """
    Implementation of a HEX8 element
    """

    ShapeName: str = "HEX8"

    def __init__(self, el_type: str, nodes: List[Node]):
        """
        Base constructor of a Hex8 element
        """
        super(Hex8, self).__init__(el_type, Hex8.ShapeName, nodes)

        if len(nodes) != self.get_num_nodes():
            raise RuntimeError(
                "You tried to created a HEX8 element with {0} nodes".format(len(nodes))
            )

    @classmethod
    def get_num_nodes(cls) -> int:
        """
        Get number of nodes of a HEX8 element

        Returns:
            Number of nodes of a Hex8 element = 8
        """
        return 8

    def get_faces(self) -> List[Quad4]:
        """
        Returns a list of faces of the Hex8 element

        Returns:
            List of faces
        """
        face_node_ids = [
            [0, 1, 2, 3],
            [0, 1, 5, 4],
            [1, 2, 6, 5],
            [2, 3, 7, 6],
            [3, 0, 4, 7],
            [4, 5, 6, 7],
        ]

        return [Quad4(None, [self.nodes[i] for i in nodes]) for nodes in face_node_ids]

    def get_edges(self) -> List[Line2]:
        """
        Returns the list of all edges

        Returns:
            List of edges
        """
        edge_node_ids = [
            [0, 1],
            [1, 2],
            [2, 3],
            [3, 0],
            [0, 4],
            [1, 5],
            [2, 6],
            [3, 7],
            [4, 5],
            [5, 6],
            [6, 7],
            [7, 4],
        ]
        return [Line2(None, [self.nodes[i] for i in nodes]) for nodes in edge_node_ids]

    @staticmethod
    def shape_fcns(xi):
        """
        Returns the value of the shape functions at the local coordinate xi
        """

        return np.array(
            [
                Line2.shape_fcns(xi[0])[0]
                * Line2.shape_fcns(xi[1])[0]
                * Line2.shape_fcns(xi[2])[0],
                Line2.shape_fcns(xi[0])[1]
                * Line2.shape_fcns(xi[1])[0]
                * Line2.shape_fcns(xi[2])[0],
                Line2.shape_fcns(xi[0])[1]
                * Line2.shape_fcns(xi[1])[1]
                * Line2.shape_fcns(xi[2])[0],
                Line2.shape_fcns(xi[0])[0]
                * Line2.shape_fcns(xi[1])[1]
                * Line2.shape_fcns(xi[2])[0],
                Line2.shape_fcns(xi[0])[0]
                * Line2.shape_fcns(xi[1])[0]
                * Line2.shape_fcns(xi[2])[1],
                Line2.shape_fcns(xi[0])[1]
                * Line2.shape_fcns(xi[1])[0]
                * Line2.shape_fcns(xi[2])[1],
                Line2.shape_fcns(xi[0])[1]
                * Line2.shape_fcns(xi[1])[1]
                * Line2.shape_fcns(xi[2])[1],
                Line2.shape_fcns(xi[0])[0]
                * Line2.shape_fcns(xi[1])[1]
                * Line2.shape_fcns(xi[2])[1],
            ]
        )

    @staticmethod
    def shape_fcns_derivs(xi):
        """
        Returns the value of the derivatives of the shape functions with respect to the local coordinates at the local coordinate xi

        +-                                  -+
        |  dN_1 / dxi_1   dN_2 / dxi_1, ...  |
        |  dN_1 / dxi_2   dN_2 / dxi_2, ...  |
        |  dN_1 / dxi_3   dN_2 / dxi_3, ...  |
        +-                                  -+
        """

        return np.array(
            [
                [
                    0.125 * (-1 + xi[1] + xi[2] - xi[1] * xi[2]),
                    0.125 * (+1 - xi[1] - xi[2] + xi[1] * xi[2]),
                    0.125 * (+1 + xi[1] - xi[2] - xi[1] * xi[2]),
                    0.125 * (-1 - xi[1] + xi[2] + xi[1] * xi[2]),
                    0.125 * (-1 + xi[1] - xi[2] + xi[1] * xi[2]),
                    0.125 * (+1 - xi[1] + xi[2] - xi[1] * xi[2]),
                    0.125 * (+1 + xi[1] + xi[2] + xi[1] * xi[2]),
                    0.125 * (-1 - xi[1] - xi[2] - xi[1] * xi[2]),
                ],
                [
                    0.125 * (-1 + xi[0] + xi[2] - xi[0] * xi[2]),
                    0.125 * (-1 - xi[0] + xi[2] + xi[0] * xi[2]),
                    0.125 * (+1 + xi[0] - xi[2] - xi[0] * xi[2]),
                    0.125 * (+1 - xi[0] - xi[2] + xi[0] * xi[2]),
                    0.125 * (-1 + xi[0] - xi[2] + xi[0] * xi[2]),
                    0.125 * (-1 - xi[0] - xi[2] - xi[0] * xi[2]),
                    0.125 * (+1 + xi[0] + xi[2] + xi[0] * xi[2]),
                    0.125 * (+1 - xi[0] + xi[2] - xi[0] * xi[2]),
                ],
                [
                    0.125 * (-1 + xi[1] + xi[0] - xi[0] * xi[1]),
                    0.125 * (-1 + xi[1] - xi[0] + xi[0] * xi[1]),
                    0.125 * (-1 - xi[1] - xi[0] - xi[0] * xi[1]),
                    0.125 * (-1 - xi[1] + xi[0] + xi[0] * xi[1]),
                    0.125 * (+1 - xi[1] - xi[0] + xi[0] * xi[1]),
                    0.125 * (+1 - xi[1] + xi[0] - xi[0] * xi[1]),
                    0.125 * (+1 + xi[1] + xi[0] + xi[0] * xi[1]),
                    0.125 * (+1 + xi[1] - xi[0] - xi[0] * xi[1]),
                ],
            ]
        )

    @staticmethod
    def nodal_reference_coordinates():
        return np.array(
            [
                [-1.0, -1.0, -1.0],
                [1.0, -1.0, -1.0],
                [1.0, 1.0, -1.0],
                [-1.0, 1.0, -1.0],
                [-1.0, -1.0, 1.0],
                [1.0, -1.0, 1.0],
                [1.0, 1.0, 1.0],
                [-1.0, 1.0, 1.0],
            ]
        )



import io
import os
import unittest
from typing import Callable

import lnmmeshio
import numpy as np
import yaml
from lnmmeshio import ioutils
from lnmmeshio.nodeset import LineNodeset, PointNodeset, SurfaceNodeset, VolumeNodeset
from parameterized import parameterized

script_dir = os.path.dirname(os.path.realpath(__file__))


def _yaml_unit_operation(dis: lnmmeshio.Discretization) -> lnmmeshio.Discretization:
    dummy_file = io.StringIO()
    dis.write_yaml(dummy_file)
    dummy_file.seek(0)

    return lnmmeshio.read_four_c_yaml(dummy_file)


def _legacy_dat_unit_operation(
    dis: lnmmeshio.Discretization,
) -> lnmmeshio.Discretization:
    dummy_file = io.StringIO()
    dis.write_legacy_dat(dummy_file)
    dummy_file.seek(0)

    return lnmmeshio.read_legacy_four_c_dat(dummy_file)


class TestDiscretizationIO(unittest.TestCase):
    def setUp(self):
        pass

    def test_read_dat(self):
        # read dat file
        mesh = lnmmeshio.read_mesh(os.path.join(script_dir, "data", "dummy.dat"))

        # check, whether all nodes are read correctly
        self.assertEqual(mesh.points.shape, (224, 3))

        # check, whether all tet elements were read correctly
        self.assertEqual(mesh.get_cells_type("tetra").shape, (12, 4))
        self.assertEqual(mesh.get_cells_type("tetra10").shape, (12, 10))
        self.assertEqual(mesh.cell_data["material"][0][0], 1)
        self.assertEqual(mesh.cell_data["material"][1][0], 4)
        self.assertEqual(mesh.cell_data["material"][2][0], 5)
        self.assertEqual(mesh.cell_data["material"][3][0], 6)

        # check, whether all hex elements were read correctly
        self.assertEqual(mesh.get_cells_type("hexahedron").shape, (65, 8))

    def test_read_yaml(self):
        # read dat file
        mesh = lnmmeshio.read_mesh(os.path.join(script_dir, "data", "dummy.4C.yaml"))

        # check, whether all nodes are read correctly
        self.assertEqual(mesh.points.shape, (224, 3))

        # check, whether all tet elements were read correctly
        self.assertEqual(mesh.get_cells_type("tetra").shape, (12, 4))
        self.assertEqual(mesh.get_cells_type("tetra10").shape, (12, 10))
        self.assertEqual(mesh.cell_data["material"][0][0], 1)
        self.assertEqual(mesh.cell_data["material"][1][0], 4)
        self.assertEqual(mesh.cell_data["material"][2][0], 5)
        self.assertEqual(mesh.cell_data["material"][3][0], 6)

        # check, whether all hex elements were read correctly
        self.assertEqual(mesh.get_cells_type("hexahedron").shape, (65, 8))

    def test_write(self):
        # read dat file
        mesh = lnmmeshio.read(os.path.join(script_dir, "data", "dummy.dat"), out=True)

        if not os.path.isdir(os.path.join(script_dir, "tmp")):
            os.makedirs(os.path.join(script_dir, "tmp"))

        # write same dat file
        lnmmeshio.write(os.path.join(script_dir, "tmp", "gen.dat"), mesh)

        # check, whether both files are identical
        # self.assertTrue(filecmp.cmp(os.path.join(script_dir, 'data', 'dummy.dat'),
        #    os.path.join((script_dir), 'tmp', 'gen.dat')))

        # check, whether both sections do coincide
        with open(os.path.join(script_dir, "data", "dummy.dat"), "r") as f:
            sections1 = ioutils.read_dat_sections(f)
        with open(os.path.join(script_dir, "tmp", "gen.dat"), "r") as f:
            sections2 = ioutils.read_dat_sections(f)

        self.assertListEqual(
            sorted(list(sections1.keys())), sorted(list(sections2.keys()))
        )

        for key in sections1.keys():
            self.assertListEqual(sorted(sections1[key]), sorted(sections2[key]))

    def test_write_yaml(self):
        # read dat file
        mesh = lnmmeshio.read(
            os.path.join(script_dir, "data", "dummy.4C.yaml"), out=True
        )

        if not os.path.isdir(os.path.join(script_dir, "tmp")):
            os.makedirs(os.path.join(script_dir, "tmp"))

        # write same dat file
        lnmmeshio.write(os.path.join(script_dir, "tmp", "gen.4C.yaml"), mesh)

        # check, whether both files are identical
        # self.assertTrue(filecmp.cmp(os.path.join(script_dir, 'data', 'dummy.dat'),
        #    os.path.join((script_dir), 'tmp', 'gen.dat')))

        # check, whether both sections do coincide
        with open(os.path.join(script_dir, "data", "dummy.4C.yaml"), "r") as f:
            sections1 = yaml.safe_load(f)
        with open(os.path.join(script_dir, "tmp", "gen.4C.yaml"), "r") as f:
            sections2 = yaml.safe_load(f)

        self.assertListEqual(
            sorted(list(sections1.keys())), sorted(list(sections2.keys()))
        )

        for key in sections1.keys():
            self.assertListEqual(sorted(sections1[key]), sorted(sections2[key]))

    @parameterized.expand([("dummy.dat",), ("dummy.4C.yaml",)])
    def test_read_new(self, file_name):
        # read discretization
        disc = lnmmeshio.read(os.path.join(script_dir, "data", file_name), out=True)

        disc.compute_ids(zero_based=False)

        self.assertEqual(len(disc.nodes), 224)
        self.assertEqual(len(disc.elements.structure), 89)

        self.assertListEqual([ns.id for ns in disc.nodes[0].surfacenodesets], [1, 13])
        self.assertListEqual([ns.id for ns in disc.nodes[47].surfacenodesets], [9])
        self.assertListEqual([ns.id for ns in disc.nodes[147].surfacenodesets], [6])

    @parameterized.expand([(_yaml_unit_operation,), (_legacy_dat_unit_operation,)])
    def test_write_new(
        self,
        unit_operation: Callable[[lnmmeshio.Discretization], lnmmeshio.Discretization],
    ):
        # build dummy discretization
        d: lnmmeshio.Discretization = lnmmeshio.Discretization()
        d.nodes = [
            lnmmeshio.Node(np.array([0.0, 0.0, 0.0])),
            lnmmeshio.Node(np.array([1.0, 0.0, 0.0])),
            lnmmeshio.Node(np.array([0.0, 1.0, 0.0])),
            lnmmeshio.Node(np.array([0.0, 0.0, 1.0])),
        ]
        for i in range(0, 4):
            d.nodes[i].fibers[lnmmeshio.Fiber.TypeFiber1] = lnmmeshio.Fiber(
                np.array([1.0, 0.0, 0.0])
            )
            d.nodes[i].fibers[lnmmeshio.Fiber.TypeFiber2] = lnmmeshio.Fiber(
                np.array([0.0, 1.0, 0.0])
            )

        d.pointnodesets.append(PointNodeset(1))
        d.linenodesets.append(LineNodeset(1))
        d.surfacenodesets.append(SurfaceNodeset(1))
        d.volumenodesets.append(VolumeNodeset(1))

        d.pointnodesets[0].add_node(d.nodes[0])

        # dline
        for i in range(0, 2):
            d.linenodesets[0].add_node(d.nodes[i])

        # dsurf
        for i in range(0, 3):
            d.surfacenodesets[0].add_node(d.nodes[i])

        # dvol
        for i in range(0, 4):
            d.volumenodesets[0].add_node(d.nodes[i])

        d.elements.structure = [lnmmeshio.Element("SOLIDSCATRA", "TET4", d.nodes)]
        d.elements.structure[0].options = {
            "MAT": 1,
            "KINEM": "nonlinear",
            "TYPE": "Std",
        }

        d_new = unit_operation(d)
        d_new.compute_ids(zero_based=True)

        self.assertEqual(len(d_new.nodes), 4)
        self.assertEqual(len(d_new.elements.structure), 1)

        self.assertAlmostEqual(
            np.linalg.norm(d_new.nodes[0].coords - np.array([0, 0, 0])), 0.0
        )
        self.assertAlmostEqual(
            np.linalg.norm(d_new.nodes[1].coords - np.array([1, 0, 0])), 0.0
        )
        self.assertAlmostEqual(
            np.linalg.norm(d_new.nodes[2].coords - np.array([0, 1, 0])), 0.0
        )
        self.assertAlmostEqual(
            np.linalg.norm(d_new.nodes[3].coords - np.array([0, 0, 1])), 0.0
        )

        for i in range(0, 4):
            self.assertAlmostEqual(
                np.linalg.norm(
                    d_new.nodes[i].fibers[lnmmeshio.Fiber.TypeFiber1].fiber
                    - np.array([1, 0, 0])
                ),
                0.0,
            )
            self.assertAlmostEqual(
                np.linalg.norm(
                    d_new.nodes[i].fibers[lnmmeshio.Fiber.TypeFiber2].fiber
                    - np.array([0, 1, 0])
                ),
                0.0,
            )

        self.assertListEqual([n.id for n in d_new.pointnodesets], [0])
        self.assertListEqual(
            [sorted([n.id for n in nds]) for nds in d_new.linenodesets], [[0, 1]]
        )
        self.assertListEqual(
            [sorted([n.id for n in nds]) for nds in d_new.surfacenodesets], [[0, 1, 2]]
        )
        self.assertListEqual(
            [sorted([n.id for n in nds]) for nds in d_new.volumenodesets],
            [[0, 1, 2, 3]],
        )

        self.assertEqual(d_new.elements.structure[0].type, "SOLIDSCATRA")
        self.assertEqual(d_new.elements.structure[0].shape, "TET4")
        self.assertListEqual(d_new.elements.structure[0].nodes, d_new.nodes)
        self.assertEqual(d_new.elements.structure[0].options["MAT"], "1")
        self.assertEqual(d_new.elements.structure[0].options["KINEM"], "nonlinear")
        self.assertEqual(d_new.elements.structure[0].options["TYPE"], "Std")

    def test_io_utils_text_fill(self):
        # test text_fill
        self.assertEqual(
            lnmmeshio.ioutils.text_fill("TEST", 10, "-", minimum=10, fill_left=True),
            "----------TEST",
        )
        self.assertEqual(
            lnmmeshio.ioutils.text_fill("TEST", 10, "-", minimum=3, fill_left=True),
            "------TEST",
        )
        self.assertEqual(
            lnmmeshio.ioutils.text_fill("TEST2", 12, "-", minimum=3), "TEST2-------"
        )

    def test_io_utils_write_title(self):
        dummy_file = io.StringIO()

        # test write_title
        lnmmeshio.ioutils.write_title(dummy_file, "TITLE")
        self.assertEqual(
            dummy_file.getvalue(),
            "--------------------------------------------------------------------TITLE\n",
        )
        dummy_file.truncate(0)
        dummy_file.seek(0)

        lnmmeshio.ioutils.write_title(
            dummy_file,
            "THIS IS A VERY LONG TITLE WHICH TAKES MORE SPACE THAN ALLOWED LETS SEE HOW",
        )
        self.assertEqual(
            dummy_file.getvalue(),
            "---THIS IS A VERY LONG TITLE WHICH TAKES MORE SPACE THAN ALLOWED LETS SEE HOW\n",
        )
        dummy_file.truncate(0)
        dummy_file.seek(0)

        lnmmeshio.ioutils.write_title(dummy_file, "TITLE", newline=False)
        self.assertEqual(
            dummy_file.getvalue(),
            "--------------------------------------------------------------------TITLE",
        )
        dummy_file.truncate(0)
        dummy_file.seek(0)

    def test_io_utils_write_option(self):
        dummy_file = io.StringIO()
        # test write_option
        lnmmeshio.ioutils.write_option(dummy_file, "KEY", "VALUE")
        self.assertEqual(
            dummy_file.getvalue(), "KEY                             VALUE\n"
        )
        dummy_file.truncate(0)
        dummy_file.seek(0)

        lnmmeshio.ioutils.write_option(dummy_file, "KEY", "VALUE", comment="COMMENT")
        self.assertEqual(
            dummy_file.getvalue(),
            "KEY                             VALUE                           // COMMENT\n",
        )
        dummy_file.truncate(0)
        dummy_file.seek(0)

        lnmmeshio.ioutils.write_option(dummy_file, "KEY", "VALUE", newline=False)
        self.assertEqual(dummy_file.getvalue(), "KEY                             VALUE")
        dummy_file.truncate(0)
        dummy_file.seek(0)

    def test_io_utils_write_option_list(self):
        dummy_file = io.StringIO()
        # test write_option_list
        lnmmeshio.ioutils.write_option_list(
            dummy_file,
            {"KEY0": "VAL0", "KEY1": ["VAL1.1", "VAL1.2"], "KEY2": [2.1, 2.2]},
        )
        self.assertEqual(
            dummy_file.getvalue(), "KEY0 VAL0 KEY1 VAL1.1 VAL1.2 KEY2 2.1 2.2\n"
        )
        dummy_file.truncate(0)
        dummy_file.seek(0)

        lnmmeshio.ioutils.write_option_list(
            dummy_file,
            {"KEY1": ["VAL1.1", "VAL1.2"], "KEY2": [2.1, 2.2]},
            newline=False,
        )
        self.assertEqual(dummy_file.getvalue(), "KEY1 VAL1.1 VAL1.2 KEY2 2.1 2.2")
        dummy_file.truncate(0)
        dummy_file.seek(0)

    def test_io_utils_write_comment(self):
        dummy_file = io.StringIO()
        # test write_comment
        lnmmeshio.ioutils.write_comment(dummy_file, "COMMENT")
        self.assertEqual(dummy_file.getvalue(), "// COMMENT\n")
        dummy_file.truncate(0)
        dummy_file.seek(0)

        lnmmeshio.ioutils.write_comment(dummy_file, "COMMENT", newline=False)
        self.assertEqual(dummy_file.getvalue(), "// COMMENT")
        dummy_file.truncate(0)
        dummy_file.seek(0)

    def test_io_utils_read_dat_sections(self):
        # build file
        dummy_file = io.StringIO()
        lnmmeshio.ioutils.write_title(dummy_file, "HEAD1")
        lnmmeshio.ioutils.write_option(dummy_file, "KEY1", "VAL1")
        lnmmeshio.ioutils.write_option(dummy_file, "KEY2", "VAL2")
        lnmmeshio.ioutils.write_title(dummy_file, "HEAD2")
        lnmmeshio.ioutils.write_option(dummy_file, "KEY1", "VAL1")
        lnmmeshio.ioutils.write_option(dummy_file, "KEY2", "VAL2")
        lnmmeshio.ioutils.write_option(dummy_file, "KEY3", "VAL3")
        lnmmeshio.ioutils.write_title(dummy_file, "HEAD3")
        lnmmeshio.ioutils.write_option(dummy_file, "KEY1", "VAL1")
        lnmmeshio.ioutils.write_option(dummy_file, "KEY2", "VAL2")
        lnmmeshio.ioutils.write_option(dummy_file, "KEY3", "VAL3")
        lnmmeshio.ioutils.write_option(dummy_file, "KEY4", "VAL4")

        # seek to beginning
        dummy_file.seek(0)
        sections = lnmmeshio.ioutils.read_dat_sections(dummy_file)
        self.assertListEqual(list(sections.keys()), ["", "HEAD1", "HEAD2", "HEAD3"])
        self.assertListEqual(sections[""], [])
        self.assertEqual(len(sections["HEAD1"]), 2)
        self.assertEqual(len(sections["HEAD2"]), 3)
        self.assertEqual(len(sections["HEAD3"]), 4)

    def test_io_utils_read_option_item(self):
        # build file
        dummy_file = io.StringIO()
        lnmmeshio.ioutils.write_option_list(
            dummy_file,
            {
                "KEY1": "VAL1",
                "KEY2": ["VAL2.1", "VAL2.2"],
                "KEY3": 3,
                "KEY4": [4.1, 4.2],
            },
        )

        line = dummy_file.getvalue()
        self.assertEqual(lnmmeshio.ioutils.read_option_item(line, "KEY1")[0], "VAL1")
        self.assertListEqual(
            lnmmeshio.ioutils.read_option_items(line, "KEY2", num=2)[0],
            ["VAL2.1", "VAL2.2"],
        )
        self.assertEqual(lnmmeshio.ioutils.read_option_item(line, "KEY3")[0], "3")
        self.assertListEqual(
            lnmmeshio.ioutils.read_option_items(line, "KEY4", num=2)[0], ["4.1", "4.2"]
        )

        self.assertRaises(
            RuntimeError, lambda: lnmmeshio.ioutils.read_option_item(line, "KEY5")
        )

    def test_io_utils_read_next_option(self):
        # build file
        dummy_file = io.StringIO()
        lnmmeshio.ioutils.write_option_list(
            dummy_file,
            {
                "KEY1": "VAL1",
                "KEY2": ["VAL2.1", "VAL2.2"],
                "KEY3": 3,
                "KEY4": [4.1, 4.2],
            },
        )

        line = dummy_file.getvalue()

        for i, (key, values) in enumerate(
            lnmmeshio.ioutils.read_key_values(
                line, lambda key: 2 if key == "KEY2" or key == "KEY4" else 1
            )
        ):
            self.assertEqual(key, f"KEY{i+1}")

            if i == 0:
                self.assertEqual(values, ["VAL1"])
            elif i == 1:
                self.assertListEqual(values, ["VAL2.1", "VAL2.2"])
            elif i == 2:
                self.assertListEqual(values, ["3"])
            elif i == 3:
                self.assertListEqual(values, ["4.1", "4.2"])

    def test_dummy_read(self):
        dummy_file = io.StringIO()
        lnmmeshio.ioutils.write_title(dummy_file, "TITLE", newline=True)
        lnmmeshio.ioutils.write_title(dummy_file, "TITLE", newline=True)
        dummy_file.seek(0)

        with self.assertRaises(ValueError) as _:
            lnmmeshio.ioutils.read_dat_sections(dummy_file)


if __name__ == "__main__":
    unittest.main()



from typing import IO, Dict, Optional

import numpy as np

from .ioutils import line_option_list, read_option_items


class Fiber:
    """
    Class that holds all information of fibers
    """

    # defintion of different fibers (add if more are necessary)
    TypeFiber1: str = "fiber1"
    TypeFiber2: str = "fiber2"
    TypeFiber3: str = "fiber3"
    TypeFiber4: str = "fiber4"
    TypeFiber5: str = "fiber5"
    TypeFiber6: str = "fiber6"
    TypeFiber7: str = "fiber7"
    TypeFiber8: str = "fiber8"
    TypeFiber9: str = "fiber9"
    TypeCir: str = "cir"
    TypeTan: str = "tan"
    TypeRad: str = "rad"
    TypeAxi: str = "axi"

    def __init__(self, fib: np.ndarray):
        """
        Initialize fiber vector in direction of fib

        Args:
            fib: Unit vector pointing in fiber direction np.array((3))
        """
        self.fiber: np.ndarray = fib

    def get_line(self, inp_type):
        """
        Gets the corresponding line in the dat file

        Args:
            inp_type: type of fiber (defined as static variable)
        """
        ftype: Optional[str] = None

        if inp_type == Fiber.TypeFiber1:
            ftype = "FIBER1"
        elif inp_type == Fiber.TypeFiber2:
            ftype = "FIBER2"
        elif inp_type == Fiber.TypeFiber3:
            ftype = "FIBER3"
        elif inp_type == Fiber.TypeFiber4:
            ftype = "FIBER4"
        elif inp_type == Fiber.TypeFiber5:
            ftype = "FIBER5"
        elif inp_type == Fiber.TypeFiber6:
            ftype = "FIBER6"
        elif inp_type == Fiber.TypeFiber7:
            ftype = "FIBER7"
        elif inp_type == Fiber.TypeFiber8:
            ftype = "FIBER8"
        elif inp_type == Fiber.TypeFiber9:
            ftype = "FIBER9"
        elif inp_type == Fiber.TypeCir:
            ftype = "CIR"
        elif inp_type == Fiber.TypeTan:
            ftype = "TAN"
        elif inp_type == Fiber.TypeRad:
            ftype = "RAD"
        elif inp_type == Fiber.TypeAxi:
            ftype = "AXI"
        else:
            raise ValueError("Unknown fiber type {0}".format(inp_type))

        return line_option_list({ftype: self.fiber})

    def write(self, dest: IO, inp_type: str) -> None:
        """
        Writes the corresponding section in the dat file into the stream type variable dest

        Args:
            dest: stream variable where to write the fiber
            inp_type: type of fiber (defined as static variable)
        """
        dest.write(self.get_line(inp_type))

    @staticmethod
    def is_fiber_type(ftype: str) -> bool:
        if ftype == "FIBER1":
            return True
        elif ftype == "FIBER2":
            return True
        elif ftype == "FIBER3":
            return True
        elif ftype == "FIBER4":
            return True
        elif ftype == "FIBER5":
            return True
        elif ftype == "FIBER6":
            return True
        elif ftype == "FIBER7":
            return True
        elif ftype == "FIBER8":
            return True
        elif ftype == "FIBER9":
            return True
        elif ftype == "CIR":
            return True
        elif ftype == "TAN":
            return True
        elif ftype == "AXI":
            return True
        elif ftype == "RAD":
            return True

        return False

    @staticmethod
    def get_fiber_type(fstr: str) -> str:
        """
        Returns the corresponding fiber type enum from the definition in the dat file

        Args:
            fstr: Fiber name as defined in the dat file

        Returns:
            Fiber enum as defined on top of the class
        """
        if fstr == "FIBER1":
            return Fiber.TypeFiber1
        elif fstr == "FIBER2":
            return Fiber.TypeFiber2
        elif fstr == "FIBER3":
            return Fiber.TypeFiber3
        elif fstr == "FIBER4":
            return Fiber.TypeFiber4
        elif fstr == "FIBER5":
            return Fiber.TypeFiber5
        elif fstr == "FIBER6":
            return Fiber.TypeFiber6
        elif fstr == "FIBER7":
            return Fiber.TypeFiber7
        elif fstr == "FIBER8":
            return Fiber.TypeFiber8
        elif fstr == "FIBER9":
            return Fiber.TypeFiber9
        elif fstr == "CIR":
            return Fiber.TypeCir
        elif fstr == "TAN":
            return Fiber.TypeTan
        elif fstr == "AXI":
            return Fiber.TypeAxi
        elif fstr == "RAD":
            return Fiber.TypeRad

        raise RuntimeError("Unknown fiber type")

    @staticmethod
    def parse_fibers(line: str) -> Dict[str, "Fiber"]:
        """
        Parses the fibers from the line and returns a dict of fiber objects

        Args:
            line: String of the line

        Returns:
            dict with fiber type as key and fiber object as value
        """
        fibs: dict = {}
        if "FIBER1" in line:
            fibs[Fiber.TypeFiber1] = Fiber(
                np.array(
                    [float(i) for i in read_option_items(line, "FIBER1", num=3)[0]]
                )
            )

        if "FIBER2" in line:
            fibs[Fiber.TypeFiber2] = Fiber(
                np.array(
                    [float(i) for i in read_option_items(line, "FIBER2", num=3)[0]]
                )
            )

        if "FIBER3" in line:
            fibs[Fiber.TypeFiber3] = Fiber(
                np.array(
                    [float(i) for i in read_option_items(line, "FIBER3", num=3)[0]]
                )
            )

        if "FIBER4" in line:
            fibs[Fiber.TypeFiber4] = Fiber(
                np.array(
                    [float(i) for i in read_option_items(line, "FIBER4", num=3)[0]]
                )
            )

        if "FIBER5" in line:
            fibs[Fiber.TypeFiber5] = Fiber(
                np.array(
                    [float(i) for i in read_option_items(line, "FIBER5", num=3)[0]]
                )
            )

        if "FIBER6" in line:
            fibs[Fiber.TypeFiber6] = Fiber(
                np.array(
                    [float(i) for i in read_option_items(line, "FIBER6", num=3)[0]]
                )
            )

        if "FIBER7" in line:
            fibs[Fiber.TypeFiber7] = Fiber(
                np.array(
                    [float(i) for i in read_option_items(line, "FIBER7", num=3)[0]]
                )
            )

        if "FIBER8" in line:
            fibs[Fiber.TypeFiber8] = Fiber(
                np.array(
                    [float(i) for i in read_option_items(line, "FIBER8", num=3)[0]]
                )
            )

        if "FIBER9" in line:
            fibs[Fiber.TypeFiber9] = Fiber(
                np.array(
                    [float(i) for i in read_option_items(line, "FIBER9", num=3)[0]]
                )
            )

        if "CIR" in line:
            fibs[Fiber.TypeCir] = Fiber(
                np.array([float(i) for i in read_option_items(line, "CIR", num=3)[0]])
            )

        if "TAN" in line:
            fibs[Fiber.TypeTan] = Fiber(
                np.array([float(i) for i in read_option_items(line, "TAN", num=3)[0]])
            )

        if "RAD" in line:
            fibs[Fiber.TypeRad] = Fiber(
                np.array([float(i) for i in read_option_items(line, "RAD", num=3)[0]])
            )

        if "AXI" in line:
            fibs[Fiber.TypeAxi] = Fiber(
                np.array([float(i) for i in read_option_items(line, "AXI", num=3)[0]])
            )

        return fibs



import math
from typing import List

import numpy as np

from ..node import Node
from .element import ElementTet
from .line3 import Line3
from .tri6 import Tri6


"""
Implementation of a tet4 element
"""


class Tet10(ElementTet):
    ShapeName: str = "TET10"

    """
    Base constructor of a tet10 element
    """

    def __init__(self, el_type: str, nodes: List[Node]):
        super(Tet10, self).__init__(el_type, Tet10.ShapeName, nodes)

        if len(nodes) != self.get_num_nodes():
            raise RuntimeError(
                "You tried to created a TET10 element with {0} nodes".format(len(nodes))
            )

    @classmethod
    def get_num_nodes(cls) -> int:
        """
        Get number of nodes of a TET10 element

        Returns:
            Number of nodes of a TET10 element = 10
        """
        return 10

    """
    Returns a list of faces of the TET10 element

    Returns:
        List of faces
    """

    def get_faces(self) -> List[Tri6]:
        face_node_ids = [
            [0, 1, 3, 4, 8, 7],
            [1, 2, 3, 5, 9, 8],
            [2, 0, 3, 6, 7, 9],
            [0, 2, 1, 6, 5, 4],
        ]

        return [Tri6(None, [self.nodes[i] for i in nodes]) for nodes in face_node_ids]

    """
    Returns the list of all edges

    Returns:
        List of edges
    """

    def get_edges(self) -> List[Line3]:
        edge_node_ids = [
            [0, 1, 4],
            [1, 2, 5],
            [2, 0, 6],
            [0, 3, 7],
            [1, 3, 8],
            [2, 3, 9],
        ]
        return [Line3(None, [self.nodes[i] for i in nodes]) for nodes in edge_node_ids]

    def get_xi(self, x):
        """
        Returns the local variables from given global variables

        Args:
            x: Global variable

        Returns:
            Local variables (xi)
        """
        coords = np.transpose(np.array([n.coords for n in self.nodes]))
        raise NotImplementedError("This is currently not implemented for all elements")

    @staticmethod
    def shape_fcns(xi):
        """
        Returns the value of the shape functions at the local coordinate xi
        """
        r = xi[0]
        s = xi[1]
        t = xi[2]
        u = 1.0 - r - s - t
        return np.array(
            [
                u * (2.0 * u - 1.0),
                r * (2.0 * r - 1.0),
                s * (2.0 * s - 1.0),
                t * (2.0 * t - 1.0),
                4.0 * r * u,
                4.0 * r * s,
                4.0 * s * u,
                4.0 * t * u,
                4.0 * r * t,
                4.0 * s * t,
            ]
        )

    @staticmethod
    def shape_fcns_derivs(xi):
        """
        Returns the value of the derivatives of the shape functions with respect to the local coordinates at the local coordinate xi

        +-                                  -+
        |  dN_1 / dxi_1   dN_2 / dxi_1, ...  |
        |  dN_1 / dxi_2   dN_2 / dxi_2, ...  |
        |  dN_1 / dxi_3   dN_2 / dxi_3, ...  |
        +-                                  -+
        """
        u = 1 - xi[0] - xi[1] - xi[2]
        return np.array(
            [
                [
                    -4 * u + 1,
                    4 * xi[0] - 1,
                    0,
                    0,
                    4 * (u - xi[0]),
                    4 * xi[1],
                    -4 * xi[1],
                    -4 * xi[2],
                    4 * xi[2],
                    0,
                ],
                [
                    -4 * u + 1,
                    0,
                    4 * xi[1] - 1,
                    0,
                    -4 * xi[0],
                    4 * xi[0],
                    4 * (u - xi[1]),
                    -4 * xi[2],
                    0,
                    4 * xi[2],
                ],
                [
                    -4 * u + 1,
                    0,
                    0,
                    4 * xi[2] - 1,
                    -4 * xi[0],
                    0,
                    -4 * xi[1],
                    4 * (u - xi[2]),
                    4 * xi[0],
                    4 * xi[1],
                ],
            ]
        )

    @staticmethod
    def nodal_reference_coordinates():
        return np.array(
            [
                [0.0, 0.0, 0.0],
                [1.0, 0.0, 0.0],
                [0.0, 1.0, 0.0],
                [0.0, 0.0, 1.0],
                [0.5, 0.0, 0.0],
                [0.5, 0.5, 0.0],
                [0.0, 0.5, 0.0],
                [0.0, 0.0, 0.5],
                [0.5, 0.0, 0.5],
                [0.0, 0.5, 0.5],
            ]
        )



import os
import unittest
from typing import List

import lnmmeshio
import numpy as np

script_dir = os.path.dirname(os.path.realpath(__file__))


class TestPyvista(unittest.TestCase):

    def setUp(self):
        pass

    def test_to_pyvista(self):
        dis = lnmmeshio.read(os.path.join(script_dir, "data", "dummy2.dat"))

        # convert to pyvista
        mesh = lnmmeshio.to_pyvista(dis)

        np.testing.assert_equal(mesh.cells, np.array([4, 0, 1, 2, 3]))
        np.testing.assert_equal(
            mesh.points,
            np.array(
                [[0.0, 0.0, 0.0], [1.0, 0.0, 0.0], [0.0, 1.0, 0.0], [0.0, 0.0, 1.0]]
            ),
        )
